# test_server.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2009 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

"""Tests for the server module."""
from __future__ import absolute_import, division, print_function
from future import standard_library

standard_library.install_aliases()  # noqa: E402

import gc
import logging
import socket
import sys
import time
import unittest
import weakref

from collections import defaultdict
from concurrent.futures import Future
from functools import partial, wraps

import _thread
import mock
import tornado.testing

from future.utils import PY2, PY3
from tornado import gen

import katcp
from katcp import __version__, kattypes
from katcp.core import FailReply
from katcp.testutils import (
    AsyncDeviceTestServer,
    BlockingTestClient,
    ClientConnectionTest,
    DeviceTestServer,
    TestLogHandler,
    TestUtilMixin,
    WaitingMock,
    handle_mock_req,
    mock_req,
    start_thread_with_cleanup,
)


log_handler = TestLogHandler()
logging.getLogger("katcp").addHandler(log_handler)
logger = logging.getLogger(__name__)

NUM_HELP_MESSAGES = 18  # Number of requests on DeviceTestServer


class test_ClientConnection(unittest.TestCase):
    def test_init(self):
        # Test that the ClientConnection methods are correctly bound to the
        # server methods

        # Check standard async inform
        server = mock.Mock(spec=katcp.server.KATCPServer(mock.Mock(), "", 0))
        raw_socket = "raw_socket"
        DUT = katcp.server.ClientConnection(server, raw_socket)
        inf_arg = katcp.Message.inform("infarg")
        DUT.inform(inf_arg)
        server.send_message.assert_called_once_with(raw_socket, inf_arg)

        # Check reply-inform
        server.send_message.reset_mock()
        mid = b"5"
        rif_req = katcp.Message.request("rif", mid=mid)
        rif_inf = katcp.Message.inform("rif")
        # Double-check that the mid's don't match before the call
        self.assertNotEqual(rif_inf.mid, mid)
        DUT.reply_inform(rif_inf, rif_req)
        # Check correct message sent
        server.send_message.assert_called_once_with(raw_socket, rif_inf)
        # Check that mid is copied over
        self.assertEqual(rif_inf.mid, mid)

        # Check that an assert is raised the original request and inform names don't match
        server.send_message.reset_mock()
        rif_req = katcp.Message.request("riffy")
        with self.assertRaises(Exception):
            DUT.reply_inform(rif_inf, rif_req)
        self.assertFalse(server.send_message.called)

        # Check reply
        server.send_message.reset_mock()
        mid = b"7"
        rep_req = katcp.Message.request("rep-req", mid=mid)
        rep_rep = katcp.Message.reply("rep-req")
        # Double-check that the mid's don't match before the call
        self.assertNotEqual(rep_rep.mid, mid)
        DUT.reply(rep_rep, rep_req)
        # Check correct message sent
        server.send_message.assert_called_once_with(raw_socket, rep_rep)
        # Check that the mid was copied over
        self.assertEqual(rep_rep.mid, mid)

        # Check that an assert is raised the original request and reply names don't match
        server.send_message.reset_mock()
        rep_rep = katcp.Message.request("reppy")
        with self.assertRaises(Exception):
            DUT.reply(rep_rep, rep_req)
        self.assertFalse(server.send_message.called)

        # Check that mass_inform works
        DUT.mass_inform(katcp.Message.inform("blahh"))
        server.mass_send_message.assert_called_once_with(katcp.Message.inform("blahh"))


class test_ClientRequestConnection(unittest.TestCase):
    def setUp(self):
        self.client_connection = mock.Mock()
        self.req_msg = katcp.Message.request("test-request", "parm1", "parm2", mid=42)
        self.DUT = katcp.server.ClientRequestConnection(
            self.client_connection, self.req_msg
        )

    def test_inform(self):
        arguments = (b"inf1", b"inf2")
        self.DUT.inform(*arguments)
        self.assertEqual(self.client_connection.inform.call_count, 1)
        (inf_msg,), kwargs = self.client_connection.inform.call_args
        self.assertSequenceEqual(inf_msg.arguments, arguments)
        self.assertEqual(inf_msg.name, "test-request")
        self.assertEqual(inf_msg.mid, b"42")
        self.assertEqual(inf_msg.mtype, katcp.Message.INFORM)

    def test_reply(self):
        arguments = (b"inf1", b"inf2")
        self.DUT.reply(*arguments)
        self.assertEqual(self.client_connection.reply.call_count, 1)
        (rep_msg, req_msg), kwargs = self.client_connection.reply.call_args
        self.assertIs(req_msg, self.req_msg)
        self.assertSequenceEqual(rep_msg.arguments, arguments)
        self.assertEqual(rep_msg.name, "test-request")
        self.assertEqual(rep_msg.mid, b"42")
        self.assertEqual(rep_msg.mtype, katcp.Message.REPLY)
        # Test that we can't reply twice
        with self.assertRaises(RuntimeError):
            self.DUT.reply(*arguments)

    def test_reply_with_msg(self):
        rep_msg = katcp.Message.reply("test-request", "inf1", "inf2")
        self.DUT.reply_with_message(rep_msg.copy())
        self.assertEqual(self.client_connection.reply.call_count, 1)
        (actual_rep_msg, req_msg), kwargs = self.client_connection.reply.call_args
        self.assertIs(req_msg, self.req_msg)
        self.assertEqual(actual_rep_msg, rep_msg)
        # Test that we can't reply twice
        with self.assertRaises(RuntimeError):
            self.DUT.reply_with_message(rep_msg)

    def test_reply_message(self):
        arguments = (b"inf1", b"inf2")
        rep_msg = self.DUT.make_reply(*arguments)
        self.assertSequenceEqual(rep_msg.arguments, arguments)
        self.assertEqual(rep_msg.name, "test-request")
        self.assertEqual(rep_msg.mid, b"42")
        self.assertEqual(rep_msg.mtype, katcp.Message.REPLY)


class TestDeviceServerV4(unittest.TestCase, TestUtilMixin):
    class DeviceTestServerV4(DeviceTestServer):
        ## Protocol versions and flags for a katcp v4 server
        PROTOCOL_INFO = katcp.ProtocolFlags(4, 0, {katcp.ProtocolFlags.MULTI_CLIENT})

    def setUp(self):
        self.server = self.DeviceTestServerV4("", 0)

    def test_log(self):
        self.server.mass_inform = mock.Mock()
        self.server.log.warn("A warning", timestamp=1234)
        self.assertEqual(self.server.mass_inform.call_count, 1)
        (msg,), _ = self.server.mass_inform.call_args
        level, timestamp, name, log_message = msg.arguments
        self.assertEqual(msg.name, "log")
        self.assertIs(msg.mid, None)
        # Timestamp should be in miliseconds
        self.assertEqual(timestamp, b"1234000")
        self.assertIn(b"A warning", log_message)

    def test_on_client_connect(self):
        fake_sock = mock.Mock()
        conn = katcp.server.ClientConnection(self.server._server, fake_sock)
        mock_conn = mock.Mock(spec=conn)
        self.server.BUILD_INFO = ("buildy", 1, 2, "g")
        self.server.VERSION_INFO = ("deviceapi", 5, 6)
        # Hack around ioloop thread asserts
        self.server._server.ioloop_thread_id = _thread.get_ident()
        # Test call
        self.server.on_client_connect(mock_conn)
        # we are expecting 2 inform messages
        no_msgs = 2
        self.assertEqual(mock_conn.inform.call_count, no_msgs)
        # Get all the inform messages
        msgs = [str(call[0][0]) for call in mock_conn.inform.call_args_list]
        self._assert_msgs_equal(
            msgs, (r"#version deviceapi-5.6", r"#build-state buildy-1.2g")
        )

    def test_sensor_sampling(self):
        start_thread_with_cleanup(self, self.server)
        s = katcp.Sensor.boolean("a-sens")
        s.set(1234, katcp.Sensor.NOMINAL, True)
        self.server.add_sensor(s)
        self.server._send_message = WaitingMock()
        self.server.wait_running(timeout=1.0)
        self.assertTrue(self.server.running())
        self.server._strategies = defaultdict(lambda: {})
        req = mock_req("sensor-sampling", "a-sens", "event")
        self.server.request_sensor_sampling(req, req.msg).result(timeout=1)
        inf = req.client_connection.inform
        inf.assert_wait_call_count(count=1)
        (inf_msg,) = inf.call_args[0]
        self._assert_msgs_equal(
            [inf_msg], (r"#sensor-status 1234000 1 a-sens nominal 1",)
        )
        req = mock_req("sensor-sampling", "a-sens", "period", 1000)
        self.server.request_sensor_sampling(req, req.msg).result()
        client = req.client_connection
        strat = self.server._strategies[client][s]
        # Test that the periodic update period is converted to seconds
        self.assertEqual(strat._period, 1.0)
        # test that parameters returned by the request matches v4 format.
        #
        # We need to pass in the same client_conn as used by the previous
        # request since strategies are bound to specific connections
        req = mock_req("sensor-sampling", "a-sens", client_conn=client)
        reply = self.server.request_sensor_sampling(req, req.msg).result()
        self._assert_msgs_equal([reply], ["!sensor-sampling ok a-sens period 1000"])
        # event-rate is not an allowed v4 strategy
        with self.assertRaises(FailReply):
            self.server.request_sensor_sampling(
                req,
                katcp.Message.request(
                    "sensor-sampling", "a-sens", "event-rate", 1000, 2000
                ),
            ).result()
        # differential-rate is not an allowed v4 strategy
        with self.assertRaises(FailReply):
            self.server.request_sensor_sampling(
                req,
                katcp.Message.request(
                    "sensor-sampling", "a-sens", "differential-rate", 1, 1000, 2000
                ),
            ).result()

    def test_sensor_value(self):
        s = katcp.Sensor.boolean("a-sens")
        s.set(1234, katcp.Sensor.NOMINAL, True)
        self.server.add_sensor(s)
        client_conn = ClientConnectionTest()
        self.server.handle_message(
            client_conn, katcp.Message.request("sensor-value", "a-sens")
        )
        self._assert_msgs_equal(
            client_conn.messages,
            ["#sensor-value 1234000 1 a-sens nominal 1", "!sensor-value ok 1"],
        )

    def test_excluded_default_handlers(self):
        """
        Test that default handers from higher KATCP versions are not included

        """
        self.assertNotIn("request-timeout-hint", self.server._request_handlers)
        self.assertNotIn("version-list", self.server._request_handlers)


class TestDeviceServerV4Async(TestDeviceServerV4):
    class DeviceTestServerV4(DeviceTestServer):
        ## Protocol versions and flags for a katcp v4 server
        PROTOCOL_INFO = katcp.ProtocolFlags(4, 0, {katcp.ProtocolFlags.MULTI_CLIENT})

        def __init__(self, *args, **kwargs):
            super(TestDeviceServerV4Async.DeviceTestServerV4, self).__init__(
                *args, **kwargs
            )
            self.set_concurrency_options(thread_safe=False, handler_thread=False)


class TestVersionCompatibility(unittest.TestCase):
    def test_wrong_version(self):
        class DeviceTestServerWrong(DeviceTestServer):
            ## Protocol versions and flags for a katcp v4 server
            PROTOCOL_INFO = katcp.ProtocolFlags(
                3, 0, {katcp.ProtocolFlags.MULTI_CLIENT}
            )

        # Only major versions 4 and 5 are supported
        with self.assertRaises(ValueError):
            DeviceTestServerWrong("", 0)
        DeviceTestServerWrong.PROTOCOL_INFO.major = 6
        with self.assertRaises(ValueError):
            DeviceTestServerWrong("", 0)


katcp_version = __version__


class test_DeviceServer(unittest.TestCase, TestUtilMixin):

    expected_connect_messages = (
        # Will have to be updated if the default KATCP protocol
        # spec version changes
        r"#version-connect katcp-protocol 5.0-IM",
        r"#version-connect katcp-library katcp-python-%s" % katcp_version,
        r"#version-connect katcp-device deviceapi-5.6 buildy-1.2g",
    )

    def setUp(self):
        self.server = DeviceTestServer("", 0)

    def test_on_client_connect(self):
        fake_sock = mock.Mock()
        mock_conn = mock.Mock(
            spec=katcp.server.ClientConnection(self.server._server, fake_sock)
        )
        self.server.BUILD_INFO = ("buildy", 1, 2, "g")
        self.server.VERSION_INFO = ("deviceapi", 5, 6)
        # Hack around ioloop thread asserts
        self.server._server.ioloop_thread_id = _thread.get_ident()
        # Test call
        self.server.on_client_connect(mock_conn)
        # we are expecting 3 inform messages
        no_msgs = 3
        self.assertEqual(mock_conn.inform.call_count, no_msgs)
        # Get all the messages sent to _send_message
        msgs = [str(call[0][0]) for call in mock_conn.inform.call_args_list]
        self._assert_msgs_equal(msgs, self.expected_connect_messages)

    def test_sensor_sampling(self):
        start_thread_with_cleanup(self, self.server)
        s = katcp.Sensor.float("a-sens")
        s.set(1234, katcp.Sensor.NOMINAL, 5545.5)
        self.server.add_sensor(s)
        self.server._send_message = WaitingMock()
        self.server.wait_running(timeout=1.0)
        self.assertTrue(self.server.running())

        # We need to pass in the same client_conn for all requests
        # since strategies are bound to specific connections.
        client = WaitingMock()
        self.server._strategies[client] = {}

        req = mock_req("sensor-sampling", "a-sens", "event", client_conn=client)
        self.server.request_sensor_sampling(req, req.msg).result(timeout=1)
        inf = req.client_connection.inform
        inf.assert_wait_call_count(count=1)
        (inf_msg,) = inf.call_args[0]
        self._assert_msgs_equal(
            [inf_msg], (r"#sensor-status 1234.000000 1 a-sens nominal 5545.5",)
        )

        req = mock_req("sensor-sampling", "a-sens", "period", 1.5, client_conn=client)
        self.server.request_sensor_sampling(req, req.msg).result()
        req = mock_req("sensor-sampling", "a-sens", client_conn=client)
        reply = self.server.request_sensor_sampling(req, req.msg).result()
        self._assert_msgs_equal([reply], ["!sensor-sampling ok a-sens period 1.5"])

        req = mock_req(
            "sensor-sampling", "a-sens", "differential", 1.5, client_conn=client
        )
        reply = self.server.request_sensor_sampling(req, req.msg).result()
        self._assert_msgs_equal([reply], ["!sensor-sampling ok a-sens differential 1.5"])

        req = mock_req(
            "sensor-sampling", "a-sens", "event-rate", 1.5, 2.5, client_conn=client
        )
        reply = self.server.request_sensor_sampling(req, req.msg).result()
        self._assert_msgs_equal(
            [reply], ["!sensor-sampling ok a-sens event-rate 1.5 2.5"]
        )

        req = mock_req(
            "sensor-sampling",
            "a-sens",
            "differential-rate",
            1.5,
            2.5,
            3.5,
            client_conn=client
        )
        reply = self.server.request_sensor_sampling(req, req.msg).result()
        self._assert_msgs_equal(
            [reply], ["!sensor-sampling ok a-sens differential-rate 1.5 2.5 3.5"]
        )

    def test_bulk_sensor_sampling_fails_if_not_supported(self):
        if not self.server.PROTOCOL_INFO.bulk_set_sensor_sampling:
            start_thread_with_cleanup(self, self.server)
            self.server._send_message = WaitingMock()
            self.server.wait_running(timeout=1.0)
            self.assertTrue(self.server.running())

            req = mock_req("sensor-sampling", "an.int,a.float", "period", 1.5)
            with self.assertRaises(FailReply):
                self.server.request_sensor_sampling(req, req.msg).result()

    def test_request_sensor_sampling_clear(self):
        self.server.clear_strategies = mock.Mock()
        start_thread_with_cleanup(self, self.server, start_timeout=1)
        client_connection = ClientConnectionTest()
        self.server.ioloop.make_current()
        tf = self.server.handle_message(
            client_connection, katcp.Message.request("sensor-sampling-clear")
        )
        # tf may be a tornado (not thread-safe) future, so we wrap it in a thread-safe
        # future object
        f = Future()
        self.server.ioloop.add_callback(gen.chain_future, tf, f)
        f.result(timeout=1)
        # Ensure that the tornado future has finished running its callbacks
        self.server.sync_with_ioloop()
        self._assert_msgs_equal(client_connection.messages, ["!sensor-sampling-clear ok"])
        self.server.clear_strategies.assert_called_once_with(client_connection)

    def test_has_sensor(self):
        self.assertFalse(self.server.has_sensor("blaah"))
        self.server.add_sensor(katcp.Sensor.boolean("blaah", "blaah sens"))
        self.assertTrue(self.server.has_sensor("blaah"))

    def test_excluded_default_handlers(self):
        """
        Test that default handers from higher KATCP versions are not included

        """
        # TODO NM 2017-04-18 This should probably be removed if we do official
        # KATCP v5.1 release and make it the default server version
        self.assertNotIn("request-timeout-hint", self.server._request_handlers)


class test_DeviceServerAsync(test_DeviceServer):
    def setUp(self):
        super(test_DeviceServerAsync, self).setUp()
        self.server.set_concurrency_options(thread_safe=False, handler_thread=False)


class test_DeviceServerMemoryLeaks(unittest.TestCase):
    def test_no_memory_leak_after_usage(self):
        server = DeviceTestServer("", 0)
        wr = weakref.ref(server)

        server.start()
        server.wait_running(timeout=1)
        server.stop()
        server.join()

        # clear strong reference and check if object can be garbage collected
        server = None
        gc.collect()
        self.assertIsNone(wr())


class test_DeviceServer51(test_DeviceServer):
    """Additional tests for Version 5.1 server"""

    expected_connect_messages = (
        r"#version-connect katcp-protocol 5.1-BIMT",
        r"#version-connect katcp-library katcp-python-%s" % katcp_version,
        r"#version-connect katcp-device deviceapi-5.6 buildy-1.2g",
    )

    def setUp(self):
        class DeviceTestServer51(DeviceTestServer):
            PROTOCOL_INFO = katcp.ProtocolFlags(
                5,
                1,
                [
                    katcp.ProtocolFlags.MULTI_CLIENT,
                    katcp.ProtocolFlags.MESSAGE_IDS,
                    katcp.ProtocolFlags.REQUEST_TIMEOUT_HINTS,
                    katcp.ProtocolFlags.BULK_SET_SENSOR_SAMPLING,
                ],
            )

        self.server = DeviceTestServer51("", 0)
        self.server.set_concurrency_options(thread_safe=False, handler_thread=False)

    def test_excluded_default_handlers(self):
        pass  # No excluded default handlers for v5.1 as of yet

    def test_request_timeout_hint(self):
        req = mock_req("request-timeout-hint")
        handle_mock_req(self.server, req)
        # Check that the default test server "slow command" request has the
        # expected timeout hint of 99 seconds
        self._assert_msgs_equal([req.reply_msg], ["!request-timeout-hint ok 1"])
        self._assert_msgs_equal(
            req.inform_msgs, ["#request-timeout-hint slow-command 99.0"]
        )
        # Check that a request without a hint gives a timeout of 0 if explicitly
        # asked for
        req = mock_req("request-timeout-hint", "help")
        handle_mock_req(self.server, req)
        self._assert_msgs_equal([req.reply_msg], ["!request-timeout-hint ok 1"])
        self._assert_msgs_equal(req.inform_msgs, ["#request-timeout-hint help 0.0"])

        # Now set hint on help command and check that it is reflected. Note, in
        # Python 2 you cannot add a new attribute to an instance method. We need
        # to make a copy of the original handler function using partial and then
        # mess with that copy.
        handler_original = self.server._request_handlers["help"]
        if PY2:
            handler_updated = wraps(handler_original)(
                partial(handler_original, self.server)
            )
        else:
            handler_updated = handler_original
        self.server._request_handlers["help"] = kattypes.request_timeout_hint(25.5)(
            handler_updated
        )
        req = mock_req("request-timeout-hint", "help")
        handle_mock_req(self.server, req)
        self._assert_msgs_equal([req.reply_msg], ["!request-timeout-hint ok 1"])
        self._assert_msgs_equal(req.inform_msgs, ["#request-timeout-hint help 25.5"])
        req = mock_req("request-timeout-hint")
        handle_mock_req(self.server, req)
        # Check that the default test server "slow command" request has the
        # expected timeout hint of 99 seconds
        self._assert_msgs_equal([req.reply_msg], ["!request-timeout-hint ok 2"])
        self._assert_msgs_equal(
            req.inform_msgs,
            [
                "#request-timeout-hint help 25.5",
                "#request-timeout-hint slow-command 99.0",
            ],
        )

    def set_up_client_for_bulk_sensor_sampling(self):
        start_thread_with_cleanup(self, self.server)
        self.server._send_message = WaitingMock()
        self.server.wait_running(timeout=1.0)
        self.assertTrue(self.server.running())

        # We need to pass in the same client_conn for all requests
        # since strategies are bound to specific connections.
        client = WaitingMock()
        self.server._strategies[client] = {}
        return client

    def test_bulk_sensor_sampling_set_for_multiple_sensors(self):
        client = self.set_up_client_for_bulk_sensor_sampling()

        req = mock_req(
            "sensor-sampling", "an.int,a.float", "period", 1.5, client_conn=client
        )
        reply = self.server.request_sensor_sampling(req, req.msg).result()
        self._assert_msgs_equal(
            [reply], ["!sensor-sampling ok an.int,a.float period 1.5"]
        )

        req = mock_req("sensor-sampling", "an.int", client_conn=client)
        reply = self.server.request_sensor_sampling(req, req.msg).result()
        self._assert_msgs_equal([reply], ["!sensor-sampling ok an.int period 1.5"])

        req = mock_req("sensor-sampling", "a.float", client_conn=client)
        reply = self.server.request_sensor_sampling(req, req.msg).result()
        self._assert_msgs_equal([reply], ["!sensor-sampling ok a.float period 1.5"])

    def test_bulk_sensor_sampling_set_for_multiple_sensors_and_individually(self):
        client = self.set_up_client_for_bulk_sensor_sampling()

        req = mock_req(
            "sensor-sampling", "an.int,a.float", "period", 1.5, client_conn=client
        )
        reply = self.server.request_sensor_sampling(req, req.msg).result()
        self._assert_msgs_equal(
            [reply], ["!sensor-sampling ok an.int,a.float period 1.5"]
        )

        req = mock_req("sensor-sampling", "an.int", "period", 3.14, client_conn=client)
        reply = self.server.request_sensor_sampling(req, req.msg).result()
        self._assert_msgs_equal([reply], ["!sensor-sampling ok an.int period 3.14"])

        req = mock_req("sensor-sampling", "a.float", "period", 0.7, client_conn=client)
        reply = self.server.request_sensor_sampling(req, req.msg).result()
        self._assert_msgs_equal([reply], ["!sensor-sampling ok a.float period 0.7"])

    def test_bulk_sensor_sampling_sends_informs_for_multiple_sensors(self):
        client = self.set_up_client_for_bulk_sensor_sampling()

        req = mock_req(
            "sensor-sampling", "an.int,a.float", "event", client_conn=client
        )
        self.server.request_sensor_sampling(req, req.msg).result(timeout=1)
        inform = req.client_connection.inform
        inform.assert_wait_call_count(count=2)
        inform_messages = [call.args[0] for call in inform.call_args_list]
        self._assert_msgs_equal(
            inform_messages, (
                r"#sensor-status 12345.000000 1 an.int nominal 3",
                r"#sensor-status 12345.000000 1 a.float nominal 12.0",
            )
        )

    def test_bulk_sensor_sampling_fails_for_invalid_sensor_name(self):
        client = self.set_up_client_for_bulk_sensor_sampling()

        req = mock_req(
            "sensor-sampling", "an.int,an.invalid", "period", 2.5, client_conn=client
        )
        with self.assertRaises(FailReply):
            self.server.request_sensor_sampling(req, req.msg).result()

    def test_bulk_sensor_sampling_fails_for_empty_sensor_name(self):
        client = self.set_up_client_for_bulk_sensor_sampling()

        req = mock_req(
            "sensor-sampling", "", "period", 2.5, client_conn=client
        )
        with self.assertRaises(FailReply):
            self.server.request_sensor_sampling(req, req.msg).result()

    def test_bulk_sensor_sampling_fails_for_invalid_strategy(self):
        client = self.set_up_client_for_bulk_sensor_sampling()

        req = mock_req(
            "sensor-sampling", "an.int,a.float", "invalid-strategy", 2.5,
            client_conn=client
        )
        with self.assertRaises(FailReply):
            self.server.request_sensor_sampling(req, req.msg).result()

    def test_bulk_sensor_sampling_fails_for_valid_strategy_invalid_params(self):
        client = self.set_up_client_for_bulk_sensor_sampling()

        req = mock_req(
            "sensor-sampling", "an.int,a.float", "period", 2.5, 1.5, client_conn=client
        )
        with self.assertRaises(FailReply):
            self.server.request_sensor_sampling(req, req.msg).result()

    def test_bulk_sensor_sampling_fails_for_valid_strategy_invalid_sensor_types(self):
        client = self.set_up_client_for_bulk_sensor_sampling()

        req = mock_req(
            "sensor-sampling", "an.int,a.discrete", "differential", 2, client_conn=client
        )
        with self.assertRaises(FailReply):
            self.server.request_sensor_sampling(req, req.msg).result()

    def test_bulk_sensor_sampling_fails_for_invalid_strategy_invalid_params(self):
        client = self.set_up_client_for_bulk_sensor_sampling()

        req = mock_req(
            "sensor-sampling", "an.int,a.float", "invalid-strategy", 2.5, 1,
            client_conn=client
        )
        with self.assertRaises(FailReply):
            self.server.request_sensor_sampling(req, req.msg).result()

    def test_bulk_sensor_sampling_fails_for_no_strategy(self):
        client = self.set_up_client_for_bulk_sensor_sampling()

        req = mock_req("sensor-sampling", "an.int,a.float", client_conn=client)
        with self.assertRaises(FailReply):
            self.server.request_sensor_sampling(req, req.msg).result()

    def test_bulk_sensor_sampling_all_strategies_unchanged_for_failure(self):
        client = self.set_up_client_for_bulk_sensor_sampling()

        req = mock_req(
            "sensor-sampling", "an.int,a.discrete", "event", client_conn=client
        )
        reply = self.server.request_sensor_sampling(req, req.msg).result()
        self._assert_msgs_equal(
            [reply], ["!sensor-sampling ok an.int,a.discrete event"]
        )
        # Save copy of current strategies for comparison later
        original_strategies = dict(self.server._strategies[client])

        # Request strategy that doesn't work for the second sensor
        req = mock_req(
            "sensor-sampling", "an.int,a.discrete", "differential", 2, client_conn=client
        )
        with self.assertRaises(FailReply):
            self.server.request_sensor_sampling(req, req.msg).result()
        self.assertEqual(self.server._strategies[client], original_strategies)

class TestDeviceServerClientIntegrated(unittest.TestCase, TestUtilMixin):

    BLACKLIST = ("version-connect", "version", "build-state")

    def setUp(self):
        super(TestDeviceServerClientIntegrated, self).setUp()
        self._setup_server()
        host, port = self.server.bind_address
        self.server_addr = (host, port)

        self.client = BlockingTestClient(self, host, port)
        start_thread_with_cleanup(self, self.client, start_timeout=1)
        self.assertTrue(self.client.wait_protocol(timeout=1))

    def _setup_server(self):
        self.server = DeviceTestServer("", 0)
        start_thread_with_cleanup(self, self.server, start_timeout=1)

    def test_log(self):
        get_msgs = self.client.message_recorder(blacklist=self.BLACKLIST, replies=True)

        def tst():
            with mock.patch("katcp.server.time.time") as m_time:
                m_time.return_value = 1234
                self.server.log.error("An error")

        self.server.ioloop.add_callback(tst)

        get_msgs.wait_number(1)
        self._assert_msgs_equal(get_msgs(), [r"#log error 1234.000000 root An\_error"])

    def test_simple_connect(self):
        """Test a simple server setup and teardown with client connect."""
        get_msgs = self.client.message_recorder(blacklist=self.BLACKLIST, replies=True)
        # basic send
        self.client.request(katcp.Message.request("foo"), use_mid=False)

        # pipe-lined send
        self.client.raw_send(b"?bar-boom\r\n?baz\r")
        # broken up sends
        self.client.raw_send(b"?boo")
        self.client.raw_send(b"m arg1 arg2")
        self.client.raw_send(b"\n")

        self._assert_msgs_equal(
            get_msgs(min_number=4),
            [
                r"!foo invalid Unknown\_request.",
                r"!bar-boom invalid Unknown\_request.",
                r"!baz invalid Unknown\_request.",
                r"!boom invalid Unknown\_request.",
            ],
        )

    def test_get_sensor(self):
        self.client.get_sensor("an.int", int)
        response = self.client.get_sensor("an.int")
        self.assertEquals(response[0], "3")
        self.client.get_sensor_value("an.int", int)
        self.client.assert_sensor_equals("an.int", 3, int, status="nominal")
        self.client.assert_sensor_status_equals("an.int", "nominal")

    def test_bad_requests(self):
        """Test request failure paths in device server."""
        get_msgs = self.client.message_recorder(blacklist=self.BLACKLIST, replies=True)
        self.client.raw_send(b"bad msg\n")
        # wait for reply
        self.client.blocking_request(katcp.Message.request("watchdog"), use_mid=False)
        self._assert_msgs_like(
            get_msgs(),
            [
                (r"#log error", r"KatcpSyntaxError:" r"\_Bad\_type\_character\_'b'.\n"),
                (r"!watchdog ok", ""),
            ],
        )

    def test_slow_client(self):
        # Test that server does not choke sending messages to slow clients

        # Set max server write buffer size smaller so that it gives up earlier to make the
        # test faster
        self.server._server.MAX_WRITE_BUFFER_SIZE = 16384

        self.client.wait_protocol(1)
        slow_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        slow_sock.connect(self.server_addr)
        slow_sock.settimeout(0.01)
        # Send a bunch of request to the server, but don't read anything from
        # the server
        try:
            slow_sock.sendall(b"?help\n" * 1000000)
        except (socket.error, socket.timeout):
            pass

        t0 = time.time()
        # Request should not have taken a very long time.
        self.client.assert_request_succeeds("help", informs_count=NUM_HELP_MESSAGES)
        self.assertTrue(time.time() - t0 < 1.5)

    def test_server_ignores_informs_and_replies(self):
        """Test server ignores informs and replies."""
        get_msgs = self.client.message_recorder(blacklist=self.BLACKLIST, replies=True)
        self.client.raw_send(b"#some inform\n")
        self.client.raw_send(b"!some reply\n")

        time.sleep(0.1)

        self.assertFalse(get_msgs())

    def test_standard_requests(self):
        """Test standard request and replies."""
        get_msgs = self.client.message_recorder(blacklist=self.BLACKLIST, replies=True)
        nomid_req = partial(self.client.blocking_request, use_mid=False)
        nomid_req(katcp.Message.request("watchdog"))
        nomid_req(katcp.Message.request("restart"))
        nomid_req(katcp.Message.request("log-level"))
        nomid_req(katcp.Message.request("log-level", "trace"))
        nomid_req(katcp.Message.request("log-level", "unknown"))
        nomid_req(katcp.Message.request("help"))
        nomid_req(katcp.Message.request("help", "watchdog"))
        nomid_req(katcp.Message.request("help", "unknown-request"))
        nomid_req(katcp.Message.request("client-list"))
        nomid_req(katcp.Message.request("version-list"))
        nomid_req(katcp.Message.request("sensor-list"))
        nomid_req(katcp.Message.request("sensor-list", "an.int"))
        nomid_req(katcp.Message.request("sensor-list", "an.unknown"))
        nomid_req(katcp.Message.request("sensor-value"))
        nomid_req(katcp.Message.request("sensor-value", "an.int"))
        nomid_req(katcp.Message.request("sensor-value", "an.unknown"))
        nomid_req(katcp.Message.request("sensor-sampling", "an.int"))
        nomid_req(katcp.Message.request("sensor-sampling", "an.int", "differential", "2"))
        nomid_req(
            katcp.Message.request("sensor-sampling", "an.int", "event-rate", "2", "3")
        )
        nomid_req(katcp.Message.request("sensor-sampling"))
        nomid_req(katcp.Message.request("sensor-sampling", "an.unknown", "auto"))
        nomid_req(katcp.Message.request("sensor-sampling", "an.int", "unknown"))

        def tst():
            self.server.log.trace("trace-msg")
            self.server.log.debug("debug-msg")
            self.server.log.info("info-msg")
            self.server.log.warn("warn-msg")
            self.server.log.error("error-msg")
            self.server.log.fatal("fatal-msg")

        self.server.ioloop.add_callback(tst)
        self.assertEqual(self.server.restart_queue.get_nowait(), self.server)
        expected_msgs = [
            (r"!watchdog ok", ""),
            (r"!restart ok", ""),
            (r"!log-level ok warn", ""),
            (r"!log-level ok trace", ""),
            (r"!log-level fail Unknown\_logging\_level\_name\_'unknown'", ""),
            (
                r"#help cancel-slow-command Cancel\_slow\_command\_request,\_"
                r"resulting\_in\_it\_replying\_immediately",
                "",
            ),
            (r"#help client-list", ""),
            (r"#help decorated", ""),
            (r"#help decorated-return-exception", ""),
            (r"#help halt", ""),
            (r"#help help", ""),
            (r"#help log-level", ""),
            (r"#help new-command", ""),
            (r"#help raise-exception", ""),
            (r"#help raise-fail", ""),
            (r"#help restart", ""),
            (r"#help sensor-list", ""),
            (r"#help sensor-sampling", ""),
            (r"#help sensor-sampling-clear", ""),
            (r"#help sensor-value", ""),
            (r"#help slow-command", ""),
            (r"#help version-list", ""),
            (r"#help watchdog", ""),
            (r"!help ok %d" % NUM_HELP_MESSAGES, ""),
            (r"#help watchdog", ""),
            (r"!help ok 1", ""),
            (r"!help fail", ""),
            (r"#client-list", ""),
            (r"!client-list ok 1", ""),
            (r"#version-list katcp-protocol", ""),
            (r"#version-list katcp-library", ""),
            (r"#version-list katcp-device", ""),
            (r"!version-list ok 3", ""),
            (r"#sensor-list a.discrete A\_Discrete. \@ discrete one two three", ""),
            (r"#sensor-list a.float A\_Float. \@ float -123.4 123.4", ""),
            (
                r"#sensor-list a.floatwithzero A\_Float\_with\_zero. \@ float -123.0"
                r" 123.0",
                "",
            ),
            (r"#sensor-list an.int An\_Integer. count integer -5 5", ""),
            (r"!sensor-list ok 4", ""),
            (r"#sensor-list an.int An\_Integer. count integer -5 5", ""),
            (r"!sensor-list ok 1", ""),
            (r"!sensor-list fail", ""),
            (r"#sensor-value 12345.000000 1 a.discrete nominal one", ""),
            (r"#sensor-value 12345.000000 1 a.float nominal 12.0", ""),
            (r"#sensor-value 12345.000000 1 a.floatwithzero nominal 12.0", ""),
            (r"#sensor-value 12345.000000 1 an.int nominal 3", ""),
            (r"!sensor-value ok 4", ""),
            (r"#sensor-value 12345.000000 1 an.int nominal 3", ""),
            (r"!sensor-value ok 1", ""),
            (r"!sensor-value fail", ""),
            (r"!sensor-sampling ok an.int none", ""),
            (r"#sensor-status 12345.000000 1 an.int nominal 3", ""),
            (r"!sensor-sampling ok an.int differential 2", ""),
            (r"#sensor-status 12345.000000 1 an.int nominal 3", ""),
            (r"!sensor-sampling ok an.int event-rate 2 3", ""),
            (r"!sensor-sampling fail No\_sensor\_name\_given.", ""),
            (r"!sensor-sampling fail Unknown\_sensor\_name:\_an.unknown.", ""),
            (r"!sensor-sampling fail Unknown\_strategy\_name:\_unknown.", ""),
            (r"#log trace", r"root trace-msg"),
            (r"#log debug", r"root debug-msg"),
            (r"#log info", r"root info-msg"),
            (r"#log warn", r"root warn-msg"),
            (r"#log error", r"root error-msg"),
            (r"#log fatal", r"root fatal-msg"),
        ]
        self._assert_msgs_like(get_msgs(min_number=len(expected_msgs)), expected_msgs)

    def test_standard_requests_with_ids(self):
        """Test standard request and replies with message ids."""
        get_msgs = self.client.message_recorder(blacklist=self.BLACKLIST, replies=True)

        current_id = [0]

        def mid():
            current_id[0] += 1
            return str(current_id[0])

        def mid_req(*args):
            return katcp.Message.request(*args, mid=mid())

        self.client.request(mid_req("watchdog"))
        self.client._next_id = mid  # mock our mid generator for testing
        self.client.blocking_request(mid_req("restart"))
        self.client.request(mid_req("log-level"))
        self.client.request(mid_req("log-level", "trace"))
        self.client.request(mid_req("log-level", "unknown"))
        self.client.request(mid_req("help"))
        self.client.request(mid_req("help", "watchdog"))
        self.client.request(mid_req("help", "unknown-request"))
        self.client.request(mid_req("client-list"))
        self.client.request(mid_req("version-list"))
        self.client.request(mid_req("sensor-list"))
        self.client.request(mid_req("sensor-list", "an.int"))
        self.client.request(mid_req("sensor-list", "an.unknown"))
        self.client.request(mid_req("sensor-value"))
        self.client.request(mid_req("sensor-value", "an.int"))
        self.client.request(mid_req("sensor-value", "an.unknown"))
        self.client.blocking_request(mid_req("sensor-sampling", "an.int"))
        self.client.blocking_request(
            mid_req("sensor-sampling", "an.int", "differential", "2")
        )
        self.client.blocking_request(
            mid_req("sensor-sampling", "an.int", "event-rate", "2", "3")
        ),
        self.client.blocking_request(mid_req("sensor-sampling"))
        self.client.blocking_request(mid_req("sensor-sampling", "an.unknown", "auto"))
        self.client.blocking_request(mid_req("sensor-sampling", "an.int", "unknown"))

        def tst():
            self.server.log.trace("trace-msg")
            self.server.log.debug("debug-msg")
            self.server.log.info("info-msg")
            self.server.log.warn("warn-msg")
            self.server.log.error("error-msg")
            self.server.log.fatal("fatal-msg")

        self.server.ioloop.add_callback(tst)

        expected_msgs = [
            (r"!watchdog[1] ok", ""),
            (r"!restart[2] ok", ""),
            (r"!log-level[3] ok warn", ""),
            (r"!log-level[4] ok trace", ""),
            (r"!log-level[5] fail Unknown\_logging\_level\_name\_'unknown'", ""),
            (
                r"#help[6] cancel-slow-command Cancel\_slow\_command\_request,\_"
                r"resulting\_in\_it\_replying\_immediately",
                "",
            ),
            (r"#help[6] client-list", ""),
            (r"#help[6] decorated", ""),
            (r"#help[6] decorated-return-exception", ""),
            (r"#help[6] halt", ""),
            (r"#help[6] help", ""),
            (r"#help[6] log-level", ""),
            (r"#help[6] new-command", ""),
            (r"#help[6] raise-exception", ""),
            (r"#help[6] raise-fail", ""),
            (r"#help[6] restart", ""),
            (r"#help[6] sensor-list", ""),
            (r"#help[6] sensor-sampling", ""),
            (r"#help[6] sensor-sampling-clear", ""),
            (r"#help[6] sensor-value", ""),
            (r"#help[6] slow-command", ""),
            (r"#help[6] version-list", ""),
            (r"#help[6] watchdog", ""),
            (r"!help[6] ok %d" % NUM_HELP_MESSAGES, ""),
            (r"#help[7] watchdog", ""),
            (r"!help[7] ok 1", ""),
            (r"!help[8] fail", ""),
            (r"#client-list[9]", ""),
            (r"!client-list[9] ok 1", ""),
            (r"#version-list[10] katcp-protocol", ""),
            (r"#version-list[10] katcp-library", ""),
            (r"#version-list[10] katcp-device", ""),
            (r"!version-list[10] ok 3", ""),
            (r"#sensor-list[11] a.discrete A\_Discrete. \@ discrete one two three", ""),
            (r"#sensor-list[11] a.float A\_Float. \@ float -123.4 123.4", ""),
            (
                r"#sensor-list[11] a.floatwithzero A\_Float\_with\_zero. \@ float -123.0"
                r" 123.0",
                "",
            ),
            (r"#sensor-list[11] an.int An\_Integer. count integer -5 5", ""),
            (r"!sensor-list[11] ok 4", ""),
            (r"#sensor-list[12] an.int An\_Integer. count integer -5 5", ""),
            (r"!sensor-list[12] ok 1", ""),
            (r"!sensor-list[13] fail", ""),
            (r"#sensor-value[14] 12345.000000 1 a.discrete nominal one", ""),
            (r"#sensor-value[14] 12345.000000 1 a.float nominal 12", ""),
            (r"#sensor-value[14] 12345.000000 1 a.floatwithzero nominal 12", ""),
            (r"#sensor-value[14] 12345.000000 1 an.int nominal 3", ""),
            (r"!sensor-value[14] ok 4", ""),
            (r"#sensor-value[15] 12345.000000 1 an.int nominal 3", ""),
            (r"!sensor-value[15] ok 1", ""),
            (r"!sensor-value[16] fail", ""),
            (r"!sensor-sampling[17] ok an.int none", ""),
            (r"#sensor-status 12345.000000 1 an.int nominal 3", ""),
            (r"!sensor-sampling[18] ok an.int differential 2", ""),
            (r"#sensor-status 12345.000000 1 an.int nominal 3", ""),
            (r"!sensor-sampling[19] ok an.int event-rate 2 3", ""),
            (r"!sensor-sampling[20] fail No\_sensor\_name\_given.", ""),
            (r"!sensor-sampling[21] fail Unknown\_sensor\_name:\_an.unknown.", ""),
            (r"!sensor-sampling[22] fail Unknown\_strategy\_name:\_unknown.", ""),
            (r"#log trace", r"root trace-msg"),
            (r"#log debug", r"root debug-msg"),
            (r"#log info", r"root info-msg"),
            (r"#log warn", r"root warn-msg"),
            (r"#log error", r"root error-msg"),
            (r"#log fatal", r"root fatal-msg"),
        ]
        self.assertEqual(self.server.restart_queue.get_nowait(), self.server)
        self._assert_msgs_like(get_msgs(min_number=len(expected_msgs)), expected_msgs)

    def test_sensor_list_regex(self):
        reply, informs = self.client.blocking_request(
            katcp.Message.request("sensor-list", "/a.*/"), use_mid=False
        )
        self._assert_msgs_equal(
            informs + [reply],
            [
                r"#sensor-list a.discrete A\_Discrete. \@ discrete one two three",
                r"#sensor-list a.float A\_Float. \@ float -123.4 123.4",
                r"#sensor-list a.floatwithzero A\_Float\_with\_zero. \@ float -123.0"
                r" 123.0",
                r"#sensor-list an.int An\_Integer. count integer -5 5",
                r"!sensor-list ok 4",
            ],
        )

        reply, informs = self.client.blocking_request(
            katcp.Message.request("sensor-list", "//"), use_mid=False
        )
        self._assert_msgs_equal(
            informs + [reply],
            [
                r"#sensor-list a.discrete A\_Discrete. \@ discrete one two three",
                r"#sensor-list a.float A\_Float. \@ float -123.4 123.4",
                r"#sensor-list a.floatwithzero A\_Float\_with\_zero. \@ float -123.0"
                r" 123.0",
                r"#sensor-list an.int An\_Integer. count integer -5 5",
                r"!sensor-list ok 4",
            ],
        )

        reply, informs = self.client.blocking_request(
            katcp.Message.request("sensor-list", "/^int/"), use_mid=False
        )
        self._assert_msgs_equal(informs + [reply], [r"!sensor-list ok 0"])

    def test_sensor_value_regex(self):
        reply, informs = self.client.blocking_request(
            katcp.Message.request("sensor-value", "/a.*/"), use_mid=False
        )
        self._assert_msgs_equal(
            informs + [reply],
            [
                r"#sensor-value 12345.000000 1 a.discrete nominal one",
                r"#sensor-value 12345.000000 1 a.float nominal 12.0",
                r"#sensor-value 12345.000000 1 a.floatwithzero nominal 12.0",
                r"#sensor-value 12345.000000 1 an.int nominal 3",
                r"!sensor-value ok 4",
            ],
        )

        reply, informs = self.client.blocking_request(
            katcp.Message.request("sensor-value", "//"), use_mid=False
        )
        self._assert_msgs_equal(
            informs + [reply],
            [
                r"#sensor-value 12345.000000 1 a.discrete nominal one",
                r"#sensor-value 12345.000000 1 a.float nominal 12.0",
                r"#sensor-value 12345.000000 1 a.floatwithzero nominal 12.0",
                r"#sensor-value 12345.000000 1 an.int nominal 3",
                r"!sensor-value ok 4",
            ],
        )

        reply, informs = self.client.blocking_request(
            katcp.Message.request("sensor-value", "/^int/"), use_mid=False
        )
        self._assert_msgs_equal(informs + [reply], [r"!sensor-value ok 0"])

    def test_client_list(self):
        reply, informs = self.client.blocking_request(
            katcp.Message.request("client-list"), use_mid=False
        )
        self.assertEqual(str(reply), "!client-list ok 1")
        self.assertEqual(len(informs), 1)
        inform = str(informs[0])
        self.assertTrue(inform.startswith("#client-list 127.0.0.1:"))
        _, addr = inform.split()
        host, port = addr.split(":")
        port = int(port)
        self.assertEqual((host, port), self.client.sockname)

    def test_halt_request(self):
        """Test halt request."""
        get_msgs = self.client.message_recorder(blacklist=self.BLACKLIST, replies=True)
        self.client.request(katcp.Message.request("halt"))
        # hack to hide re-connect exception
        self.client.connect = lambda: None
        self.server.join()

        self._assert_msgs_equal(
            get_msgs(min_number=2),
            [r"!halt[1] ok", r"#disconnect Device\_server\_shutting\_down."],
        )

    def test_bad_handlers(self):
        """Test that bad request and inform handlers are picked up."""
        try:

            class BadServer(katcp.DeviceServer):
                def request_baz(self, req, msg):
                    pass

        except AssertionError:
            pass
        else:
            self.fail("Server metaclass accepted missing request_ docstring.")

        try:

            class BadServer(katcp.DeviceServer):
                def inform_baz(self, req, msg):
                    pass

        except AssertionError:
            pass
        else:
            self.fail("Server metaclass accepted missing inform_ docstring.")

        class SortOfOkayServer(katcp.DeviceServer):
            request_bar = 1
            inform_baz = 2

        assert "bar" not in SortOfOkayServer._request_handlers
        assert "baz" not in SortOfOkayServer._inform_handlers

    def test_handler_exceptions(self):
        """Test handling of failure replies and other exceptions."""
        get_msgs = self.client.message_recorder(blacklist=self.BLACKLIST, replies=True)
        self.assertTrue(self.client.wait_protocol(timeout=1))

        self.client.request(katcp.Message.request("raise-exception"))
        self.client.request(katcp.Message.request("raise-fail"))

        time.sleep(0.1)

        self._assert_msgs_like(
            get_msgs(),
            [
                (r"!raise-exception[1] fail Traceback", ""),
                (r"!raise-fail[2] fail There\_was\_a\_problem\_with\_your\_request.", ""),
            ],
        )

    def test_decorated_handler(self):
        """Test handling with @request and @return_reply."""

        _string = "sss"
        _float = 1.5
        _int = 78

        do_raise = False
        ok_msg = katcp.Message.request("decorated", _string, _float, _int, do_raise)
        reply, informs = self.client.blocking_request(ok_msg, use_mid=False)
        self._assert_msgs_equal(
            informs + [reply],
            [
                str(katcp.Message.inform("decorated", "_string", type(_string))),
                str(katcp.Message.inform("decorated", "_float", type(_float))),
                str(katcp.Message.inform("decorated", "_int", type(_int))),
                str(katcp.Message.inform("decorated", "do_raise", type(do_raise))),
                r"!decorated ok sss 1.5 78 0",
            ],
        )

        do_raise = True
        fail_msg = katcp.Message.request("decorated", _string, _float, _int, do_raise)
        reply, informs = self.client.blocking_request(fail_msg, use_mid=False)
        self._assert_msgs_like(
            informs + [reply],
            [(r"!decorated fail Traceback", r"An\_exception\_occurred!\n")],
        )

    def test_decorated_return_exception_handler(self):
        """Test handling with @return_reply exception object."""
        msg = katcp.Message.request("decorated-return-exception")
        reply, informs = self.client.blocking_request(msg, use_mid=False)
        self._assert_msgs_like(
            informs + [reply],
            [(r"!decorated-return-exception fail An\_exception\_occurred!", "")],
        )

    def test_stop_and_restart(self):
        """Test stopping and restarting the device server."""
        # So we can wait for the client to disconnect
        self.client.notify_connected = WaitingMock()
        self.server.stop(timeout=0.1)
        self.server.join(timeout=1.0)
        self.assertFalse(self.server.running())
        # Wait for client to be disconnected
        self.client.notify_connected.assert_wait_call_count(1)
        self.assertFalse(self.client.is_connected())
        self.server.start(timeout=1.0)

    def test_bad_client_socket(self):
        """Test what happens when select is called on a dead client socket."""
        # close client stream while the server isn't looking then wait for the server to
        # notice

        stream, client_conn = list(self.server._server._connections.items())[0]
        # Wait for the client to disconnect
        self.client.notify_connected = WaitingMock()
        self.server.ioloop.add_callback(stream.close)
        # Wait for the client to be disconnected, and to connect again
        self.client.notify_connected.assert_wait_call_count(2)
        self.server.sync_with_ioloop()

        # check that client stream was removed from the KATCPServer
        self.assertTrue(
            stream not in self.server._server._connections,
            "Expected %r to not be in %r" % (stream, self.server._server._connections),
        )
        # And check that the ClientConnection object was removed from the DeviceServer
        self.assertTrue(
            client_conn not in self.server._client_conns,
            "Expected %r to not be in %r" % (client_conn, self.server._client_conns),
        )

    def test_sampling(self):
        """Test sensor sampling."""
        get_msgs = self.client.message_recorder(blacklist=self.BLACKLIST, replies=True)
        self.client.wait_protocol(timeout=1)
        self.client.request(
            katcp.Message.request("sensor-sampling", "an.int", "period", 1 / 32.0)
        )

        # Wait for the request reply and for the sensor update messages to
        # arrive. We expect update one the moment the sensor-sampling request is
        # made, then four more over 4/32. of a second, resutling in 6
        # messages. Wait 0.75 of a period longer just to be sure we get everything.

        self.assertTrue(get_msgs.wait_number(6, timeout=4.75 / 32.0))
        self.client.assert_request_succeeds("sensor-sampling", "an.int", "none")
        # Wait for reply to above request
        get_msgs.wait_number(7)
        msgs = get_msgs()
        updates = [x for x in msgs if x.name == "sensor-status"]
        others = [x for x in msgs if x.name != "sensor-status"]
        self.assertTrue(
            abs(len(updates) - 5) < 2, "Expected 5 informs, saw %d." % len(updates)
        )

        self._assert_msgs_equal(
            others,
            [
                r"!sensor-sampling[1] ok an.int period %s" % (1 / 32.0),
                r"!sensor-sampling[2] ok an.int none",
            ],
        )

        self.assertEqual(updates[0].arguments[1:], [b"1", b"an.int", b"nominal", b"3"])

        ## Now clear the strategies on this sensor
        # There should only be on connection to the server, so it should be
        # the test client
        client_conn = list(self.server._client_conns)[0]
        self.server.ioloop.add_callback(self.server.clear_strategies, client_conn)
        self.server.sync_with_ioloop()
        self.client.assert_request_succeeds(
            "sensor-sampling", "an.int", args_equal=["an.int", "none"]
        )
        # Check that we did not accidentally clobber the strategy datastructure
        # in the proccess
        self.client.assert_request_succeeds("sensor-sampling", "an.int", "period", 0.125)

    def test_assert_request_fails(self):
        self.client.assert_request_fails("sensor-sampling", "an.nonexistentsensor")
        self.client.assert_request_fails(
            "sensor-sampling",
            "an.nonexistentsensor",
            error_equals="Unknown sensor name: an.nonexistentsensor.",
        )
        self.client.assert_request_fails(
            "sensor-sampling", "an.nonexistentsensor", status_equals="fail"
        )
        with self.assertRaises(Exception):
            self.client.assert_request_fails(
                "sensor-sampling", "an.int", status_equals="fail"
            )
        with self.assertRaises(Exception):
            self.client.assert_request_fails(
                "sensor-sampling",
                "an.int",
                error_equals="Unknown sensor name: an.nonexistentsensor.",
            )

    def test_test_sensor_list(self):
        """Test that exercises test_sensor_list"""
        byte_sensors = {
            (b"a.discrete", b"A Discrete.", b"", b"discrete", b"one", b"two", b"three"),
            (b"a.float", b"A Float.", b"", b"float", b"-123.4", b"123.4"),
            (
                b"a.floatwithzero",
                b"A Float with zero.",
                b"",
                b"float",
                b"-123.0",
                b"123.0",
            ),
            (b"an.int", b"An Integer.", b"count", b"integer", b"-5", b"5"),
        }
        self.client.test_sensor_list(byte_sensors)
        self.client.test_sensor_list(byte_sensors, ignore_descriptions=True)
        str_sensors = {
            ("a.discrete", "A Discrete.", "", "discrete", "one", "two", "three"),
            ("a.float", "A Float.", "", "float", "-123.4", "123.4"),
            ("a.floatwithzero", "A Float with zero.", "", "float", "-123.0", "123.0"),
            ("an.int", "An Integer.", "count", "integer", "-5", "5"),
        }
        self.client.test_sensor_list(str_sensors)
        self.client.test_sensor_list(str_sensors, ignore_descriptions=True)

        mix_sensors = {
            ("a.discrete", "A Discrete.", "", "discrete", b"one", b"two", b"three"),
            ("a.float", b"A Float.", "", b"float", "-123.4", "123.4"),
            (b"a.floatwithzero", "A Float with zero.", b"", "float", b"-123.0", "123.0"),
            ("an.int", b"An Integer.", "count", "integer", "-5", b"5"),
        }
        self.client.test_sensor_list(mix_sensors)
        self.client.test_sensor_list(mix_sensors, ignore_descriptions=True)

    def test_assert_request_succeeds(self):
        """Test exercises assert_request_succeeds"""

        self.client.assert_request_succeeds(
            "sensor-sampling", "an.int", args_equal=["an.int", "none"]
        )
        self.client.assert_request_succeeds(
            "sensor-sampling", b"an.int", args_equal=["an.int", "none"]
        )
        self.client.assert_request_succeeds(
            "sensor-sampling", b"an.int", args_equal=[b"an.int", b"none"]
        )
        self.client.assert_request_succeeds("log-level", "debug", args_echo=True)
        levels = ["off", b"fatal", "error", b"warn", "info", "debug", "trace", "all"]
        self.client.assert_request_succeeds(
            "log-level", args_in=[[level] for level in levels]
        )
        self.client.assert_request_succeeds(
            "sensor-value",
            "an.int",
            args_equal=["1"],
            informs_args_equal=[["12345.000000", "1", "an.int", "nominal", "3"]],
        )

        with self.assertRaises(Exception):
            self.client.assert_request_succeeds(
                "sensor-value",
                "an.int",
                args_equal=["1"],
                informs_args_equal=[["12345.000000", "1", "an.int", "nominal", "3a"]],
            )

    def test_add_remove_sensors(self):
        """Test adding and removing sensors from a running device."""
        an_int = self.server._sensors["an.int"]
        self.server.remove_sensor(an_int)
        # TODO remove_sensor test that checks that everything is indeed gone
        self.server.add_sensor(an_int)
        self.test_sampling()

    def test_async_request_handler(self):
        """
        Request handlers allowing other requests to be handled before replying

        """
        # We use a coroutine running on the client's ioloop for this test
        @gen.coroutine
        def do_async_request_test(threadsafe_future, tornado_future):
            try:
                slow_wait_time = 20
                # Kick off a slow, async request
                slow_f = self.client.future_request(
                    katcp.Message.request("slow-command", slow_wait_time)
                )
                t0 = time.time()
                # Do another normal request that should reply before the slow
                # request
                reply, _ = yield self.client.future_request(
                    katcp.Message.request("watchdog")
                )
                self.assertTrue(reply.reply_ok(), "Normal request should succeed")
                # Normal request should reply quickly
                t1 = time.time()
                self.assertTrue(
                    t1 - t0 < 0.1 * slow_wait_time,
                    "The normal request should reply almost immediately",
                )
                # Slow request should still be outstanding
                self.assertFalse(
                    slow_f.done(), "slow async request should not be complete yet"
                )
                # Now cancel the slow command
                reply, _ = yield self.client.future_request(
                    katcp.Message.request("cancel-slow-command")
                )
                self.assertTrue(reply.reply_ok(), "?cancel-slow-command should succeed")
                slow_reply, _ = yield slow_f
                self.assertTrue(
                    slow_reply.reply_ok(),
                    "?slow-command should succeed after being cancelled",
                )
                t2 = time.time()
                self.assertTrue(
                    t2 - t1 < 0.1 * slow_wait_time,
                    "Slow request should be cancelled almost immediately",
                )
            except Exception as exc:
                tornado_future.set_exc_info(sys.exc_info())
                threadsafe_future.set_exception(exc)
            else:
                threadsafe_future.set_result(None)

        tornado_future = gen.Future()
        threadsafe_future = Future()
        self.client.ioloop.add_callback(
            do_async_request_test, threadsafe_future, tornado_future
        )
        try:
            threadsafe_future.result()
        except Exception:
            # Use the tornado future to get a usable traceback
            tornado_future.result()

    def test_test_help(self):
        """Basic Test to exercise test_help function
        """
        request_names = [
            "cancel-slow-command",
            "decorated",
            "decorated-return-exception",
            "new-command",
            "raise-exception",
            "raise-fail",
            "slow-command",
        ]
        self.client.test_help(request_names)

    def test_test_help_failure(self):
        """Basic Test to exercise test_help function
        failure by supplying byte strings in Python 3.
        """
        request_names = [
            b"cancel-slow-command",
            b"decorated",
            b"decorated-return-exception",
            b"new-command",
            b"raise-exception",
            b"raise-fail",
            b"slow-command",
        ]
        if PY3:
            with self.assertRaises(ValueError):
                self.client.test_help(request_names)
        # Please note that this does not fail in python2
        if PY2:
            self.client.test_help(request_names)


class TestHandlerFiltering(unittest.TestCase):
    class DeviceWithEverything(katcp.DeviceServer):
        PROTOCOL_INFO = katcp.ProtocolFlags(
            5,
            1,
            [
                katcp.ProtocolFlags.MULTI_CLIENT,
                katcp.ProtocolFlags.MESSAGE_IDS,
                katcp.ProtocolFlags.REQUEST_TIMEOUT_HINTS,
            ],
        )

        def request_simple(self, req, msg):
            """A simple request"""

        @kattypes.return_reply()
        @kattypes.minimum_katcp_version(5, 1)
        def request_version_51(self, req, msg):
            """Request KATCP v5.1"""

        @kattypes.minimum_katcp_version(5, 0)
        def request_version_5(self, req, msg):
            """Request KATCP v5.0"""

        @kattypes.request()
        @kattypes.has_katcp_protocol_flags(
            [katcp.ProtocolFlags.MULTI_CLIENT, katcp.ProtocolFlags.MESSAGE_IDS]
        )
        def request_flags(self, req):
            """Request some flags"""

        @kattypes.has_katcp_protocol_flags([katcp.ProtocolFlags.MULTI_CLIENT])
        def request_fewer_flags(self, req, msg):
            """Request with fewer flags"""

        @kattypes.request()
        @kattypes.return_reply()
        @kattypes.minimum_katcp_version(5, 1)
        @kattypes.has_katcp_protocol_flags([katcp.ProtocolFlags.MULTI_CLIENT])
        def request_version_and_flags(self, req, msg):
            """Request with version and flag requirements"""

    all_expected_handlers = frozenset(
        ["simple", "version-51", "version-5", "flags", "fewer-flags", "version-and-flags"]
    )

    def _test_handler_protocol_filters(self, DeviceCls, expected_handlers):
        """Test that handlers are filtered according to protocol requirements."""
        expected_handlers = set(expected_handlers)
        not_expected_handlers = self.all_expected_handlers - expected_handlers
        actual_device_handlers = set(DeviceCls._request_handlers.keys())

        self.assertEqual(actual_device_handlers & expected_handlers, expected_handlers)
        self.assertEqual(set(), actual_device_handlers & not_expected_handlers)

    def test_handler_protocol_filters_all(self):
        """Test handler filtering where everything should be included"""
        self._test_handler_protocol_filters(
            self.DeviceWithEverything, self.all_expected_handlers
        )

    def test_handler_protocol_filters_five_one_with_nothing(self):
        """Test handler filtering where protocol flags exclude some"""

        class DeviceVersionFiveOneWithNothing(self.DeviceWithEverything):
            PROTOCOL_INFO = katcp.ProtocolFlags(5, 1, [])

        self._test_handler_protocol_filters(
            DeviceVersionFiveOneWithNothing, ["simple", "version-5", "version-51"]
        )

    def test_handler_protocol_filters_five_one_with_multi(self):
        """Test handler filtering where protocol flags and version exclude some"""

        class DeviceVersionFiveOneWithMultiClient(self.DeviceWithEverything):
            PROTOCOL_INFO = katcp.ProtocolFlags(5, 1, [katcp.ProtocolFlags.MULTI_CLIENT])

        self._test_handler_protocol_filters(
            DeviceVersionFiveOneWithMultiClient,
            ["simple", "version-5", "version-51", "fewer-flags", "version-and-flags"],
        )

    def test_handler_protocol_filters_five(self):
        """Test handler filtering for a KATCP v5.0 device"""

        class DeviceVersionFive(self.DeviceWithEverything):
            PROTOCOL_INFO = katcp.ProtocolFlags(
                5, 0, [katcp.ProtocolFlags.MULTI_CLIENT, katcp.ProtocolFlags.MESSAGE_IDS]
            )

        self._test_handler_protocol_filters(
            DeviceVersionFive, ["simple", "version-5", "flags", "fewer-flags"]
        )

    def test_handler_protocol_filters_four(self):
        """Test handler filtering for a KATCP v4.0 device"""

        class DeviceVersionFour(self.DeviceWithEverything):
            PROTOCOL_INFO = katcp.ProtocolFlags(4, 0, [katcp.ProtocolFlags.MULTI_CLIENT])

        self._test_handler_protocol_filters(DeviceVersionFour, ["simple", "fewer-flags"])


class TestDeviceServerClientIntegratedAsync(
    tornado.testing.AsyncTestCase, TestDeviceServerClientIntegrated
):
    def _setup_server(self):
        self.server = AsyncDeviceTestServer("", 0)
        start_thread_with_cleanup(self, self.server, start_timeout=1)



# test_compat.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2020 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

"""Tests for the server module."""
from __future__ import absolute_import, division, print_function
from future import standard_library
standard_library.install_aliases()  # noqa: E402

import unittest

from katcp.compat import ensure_native_str


class test_CompatClass(unittest.TestCase):
    def test_ensure_native_str_raises_error(self):
        """Test that ensure_native_str raises an error"""
        with self.assertRaises(TypeError):
            ensure_native_str(9)
        with self.assertRaises(TypeError):
            ensure_native_str(False)



# test_katcp.py
# -*- coding: utf-8 -*-
# vim:fileencoding=utf-8 ai ts=4 sts=4 et sw=4
# Copyright 2009 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

"""Tests for the katcp utilities module.
   """
from __future__ import absolute_import, division, print_function
from future import standard_library
standard_library.install_aliases()  # noqa: E402

import logging
import unittest

import future
import tornado

import katcp

from katcp import KatcpSyntaxError, KatcpTypeError
from katcp.core import (AsyncEvent, AsyncState, Message, Sensor,
                        hashable_identity, until_some)
from katcp.testutils import TestLogHandler

log_handler = TestLogHandler()
logging.getLogger("katcp").addHandler(log_handler)


class TestMessage(unittest.TestCase):
    def test_init_basic(self):
        msg = Message(Message.REQUEST,
                      'hello', ['world', b'binary\xff\x00', 123, 4.5, True, False])
        self.assertEqual(msg.mtype, Message.REQUEST)
        self.assertEqual(msg.name, 'hello')
        self.assertEqual(msg.arguments, [
            b'world', b'binary\xff\x00', b'123', b'4.5', b'1', b'0'])
        self.assertIsNone(msg.mid)

    def test_init_mid(self):
        msg = Message(Message.REPLY, 'hello', ['world'], mid=345)
        self.assertEqual(msg.mtype, Message.REPLY)
        self.assertEqual(msg.name, 'hello')
        self.assertEqual(msg.arguments, [b'world'])
        self.assertEqual(msg.mid, b'345')

    def test_init_bad_name(self):
        self.assertRaises(ValueError, Message, Message.REPLY,
                          'underscores_bad', 'world', mid=345)
        self.assertRaises(ValueError, Message, Message.REPLY,
                          '', 'world', mid=345)
        self.assertRaises(ValueError, Message, Message.REPLY,
                          '1numberfirst', 'world', mid=345)

    def test_reply_ok(self):
        """Test reply checking."""
        self.assertEqual(Message.reply("foo", "ok").reply_ok(), True)
        self.assertEqual(Message.reply("foo", "ok", 1).reply_ok(), True)
        self.assertEqual(Message.reply("foo", "fail").reply_ok(), False)
        self.assertEqual(Message.reply("foo", "fail", "ok").reply_ok(),
                         False)
        self.assertEqual(Message.request("foo", "ok").reply_ok(), False)

    def test_request(self):
        """Test request method."""
        self.assertEqual(bytes(Message.request("foo")), b"?foo")
        self.assertEqual(bytes(Message.request("foo", mid=123)),
                         b"?foo[123]")
        self.assertEqual(bytes(Message.request("foo", "a", "b", mid=123)),
                         b"?foo[123] a b")

    def test_request_attributes(self):
        """Test request message attributes."""
        msg = Message.request('hello', 'world')
        self.assertEqual(msg.mtype, Message.REQUEST)
        self.assertEqual(msg.name, 'hello')
        self.assertEqual(msg.arguments, [b'world'])
        self.assertIsNone(msg.mid)

    def test_reply(self):
        """Test reply method."""
        self.assertEqual(bytes(Message.reply("foo")), b"!foo")
        self.assertEqual(bytes(Message.reply("foo", mid=123)), b"!foo[123]")
        self.assertEqual(bytes(Message.reply("foo", "a", "b", mid=123)),
                         b"!foo[123] a b")

    def test_reply_attributes(self):
        """Test reply message attributes."""
        msg = Message.reply('hello', 'world', mid=None)
        self.assertEqual(msg.mtype, Message.REPLY)
        self.assertEqual(msg.name, 'hello')
        self.assertEqual(msg.arguments, [b'world'])
        self.assertIsNone(msg.mid)

    def test_inform(self):
        """Test inform method."""
        self.assertEqual(bytes(Message.inform("foo")), b"#foo")
        self.assertEqual(bytes(Message.inform("foo", mid=123)),
                         b"#foo[123]")
        self.assertEqual(bytes(Message.inform("foo", "a", "b", mid=123)),
                         b"#foo[123] a b")

    def test_inform_attributes(self):
        """Test inform message attributes."""
        msg = Message.inform('hello', 'world', mid=1)
        self.assertEqual(msg.mtype, Message.INFORM)
        self.assertEqual(msg.name, 'hello')
        self.assertEqual(msg.arguments, [b'world'])
        self.assertEqual(msg.mid, b'1')

    def test_reply_to_request(self):
        req = Message.request('hello', 'world', mid=1)
        reply = Message.reply_to_request(req, 'test')
        self.assertEqual(reply.mtype, Message.REPLY)
        self.assertEqual(reply.name, 'hello')
        self.assertEqual(reply.arguments, [b'test'])
        self.assertEqual(reply.mid, b'1')

    def test_equality(self):
        class AlwaysEqual(object):
            def __eq__(self, other):
                return True

        msg = Message.inform("foo", "a", "b")
        assert msg == Message.inform("foo", "a", "b")
        assert msg != Message.request("foo", "a", "b")
        assert msg != Message.inform("bar", "a", "b")
        assert msg != Message.inform("foo", "a", "b", "c")
        assert msg != Message.reply("foo", "a", "b")
        assert msg != 3
        assert msg == AlwaysEqual()

    def test_bytes(self):
        msg = Message.request(
            'hello', u'café', b'_bin ary\xff\x00\n\r\t\\\x1b', 123, 4.5, True, False, '')
        raw = bytes(msg)
        expected = (
            b'?hello caf\xc3\xa9 _bin\\_ary\xff\\0\\n\\r\\t\\\\\\e 123 4.5 1 0 \\@')
        self.assertEqual(raw, expected)

    def test_bytes_mid(self):
        msg = Message.reply('fail', 'on fire', mid=234)
        self.assertEqual(bytes(msg), b'!fail[234] on\\_fire')

    def test_str(self):
        msg = Message.reply('fail', 'on fire', mid=234)
        self.assertEqual(str(msg), '!fail[234] on\\_fire')

        # Expect slighty different results in case of invalid UTF-8 bytes for
        # PY2's native byte string compared to PY3's unicode
        msg = Message.reply('ok', b'invalid utf-8\xff\x00')
        if future.utils.PY2:
            self.assertEqual(str(msg), '!ok invalid\_utf-8\xff\\0')
        else:
            with self.assertRaises(UnicodeDecodeError):
                str(msg)

    def test_repr(self):
        msg = Message.reply(
            'ok', 'café', b'_bin ary\xff\x00\n\r\t\\\x1b', 123, 4.5, True, False, '',
            'z'*1100)
        # storing Message.arguments as byte string results in slightly different
        # reprs for PY2 compared to PY3.
        if future.utils.PY2:
            expected = ("<Message reply ok (caf\xc3\xa9, "
                        "_bin\\_ary\xff\\0\\n\\r\\t\\\\\\e, "
                        "123, 4.5, 1, 0, , "
                        "{}..."
                        ")>".format('z'*1000))
        else:
            expected = (r"<Message reply ok (b'caf\xc3\xa9', "
                        r"b'_bin\\_ary\xff\\0\\n\\r\\t\\\\\\e', "
                        r"b'123', b'4.5', b'1', b'0', b'', "
                        r"b'{}...'"
                        r")>".format('z'*1000))
        self.assertEqual(repr(msg), expected)

    def test_bad_utf8_unicode(self):
        # Not great to have a test limited to PY3, but the 'utf-8' encoder
        # doesn't complain about this string on PY2
        if future.utils.PY3:
            msg = Message(Message.REQUEST, 'hello', [u'bad\ud83d\ude04string'])
            self.assertEqual(msg.arguments, [b'bad??string'])

    def test_stringy_arguments_to_bytes(self):
        """Test non-simple types get correctly converted to bytes."""

        class Stringy(object):
            def __str__(self):
                return "string"

        req = Message.request('hello', Stringy())
        self.assertEqual(req.arguments, [b"string"])


class TestMessageParser(unittest.TestCase):
    def setUp(self):
        self.p = katcp.MessageParser()

    def test_simple_messages(self):
        """Test simple messages."""
        m = self.p.parse(b"?foo")
        self.assertEqual(m.mtype, m.REQUEST)
        self.assertEqual(m.name, "foo")
        self.assertEqual(m.arguments, [])

        m = self.p.parse(b"#bar 123 baz 1.000e-05")
        self.assertEqual(m.mtype, m.INFORM)
        self.assertEqual(m.name, "bar")
        self.assertEqual(m.arguments, [b"123", b"baz", b"1.000e-05"])

        m = self.p.parse(b"!baz a17 goo")
        self.assertEqual(m.mtype, m.REPLY)
        self.assertEqual(m.name, "baz")
        self.assertEqual(m.arguments, [b"a17", b"goo"])

    def test_escape_sequences(self):
        """Test escape sequences."""
        m = self.p.parse(br"?foo \\\_\0\n\r\e\t\@")
        self.assertEqual(m.arguments, [b"\\ \0\n\r\x1b\t"])

        self.assertRaises(KatcpSyntaxError, self.p.parse, br"?foo \z")

        # test unescaped null
        self.assertRaises(KatcpSyntaxError, self.p.parse, b"?foo \0")

    def test_syntax_errors(self):
        """Test generation of syntax errors."""
        self.assertRaises(KatcpSyntaxError, self.p.parse, br" ?foo")
        self.assertRaises(KatcpSyntaxError, self.p.parse, br"? foo")
        self.assertRaises(KatcpSyntaxError, self.p.parse, br"?1foo")
        self.assertRaises(KatcpSyntaxError, self.p.parse, br">foo")
        self.assertRaises(KatcpSyntaxError, self.p.parse, b"!foo \\")
        self.assertRaises(KatcpSyntaxError, self.p.parse, b"!test message\n")
        self.assertRaises(KatcpSyntaxError, self.p.parse, b"")
        # invalid escapes
        self.assertRaises(KatcpSyntaxError, self.p.parse, b'!ok \\q')
        self.assertRaises(KatcpSyntaxError, self.p.parse, b'!ok q\\ other')
        # unescaped
        self.assertRaises(KatcpSyntaxError, self.p.parse, b'!ok \x1b')

    def test_message_to_string(self):
        """Test message to string round trip with escapes."""
        for m_str in [
            b"?bar",
            br"?foo \\\_\0\n\r\e\t",
        ]:
            self.assertEqual(m_str, bytes(self.p.parse(m_str)))

    def test_command_names(self):
        """Test a variety of command names."""
        m = self.p.parse(b"!baz-bar")
        self.assertEqual(m.name, "baz-bar")
        self.assertRaises(KatcpSyntaxError, self.p.parse, br"?-foo")
        self.assertRaises(KatcpSyntaxError, self.p.parse, br"?bad_name")
        self.assertRaises(KatcpSyntaxError, self.p.parse, br"? empty")

    def test_empty_params(self):
        """Test parsing messages with empty parameters."""
        m = self.p.parse(b"!foo \@")  # 1 empty parameter
        self.assertEqual(m.arguments, [b""])
        m = self.p.parse(b"!foo \@ \@")  # 2 empty parameter
        self.assertEqual(m.arguments, [b"", b""])
        m = self.p.parse(b"!foo \_ \_ \@")  # space, space, empty
        self.assertEqual(m.arguments, [b" ", b" ", b""])

    def test_whitespace(self):
        """Test parsing of whitespace between parameters."""
        m = self.p.parse(b"!baz   \@   ")  # 1 empty parameter
        self.assertEqual(m.arguments, [b""])
        m = self.p.parse(b"!baz\t\@\t\@")  # 2 empty parameter
        self.assertEqual(m.arguments, [b"", b""])
        m = self.p.parse(b"!baz\t \t\_\t\t\t \_\t\@   \t")  # space, space, \@
        self.assertEqual(m.arguments, [b" ", b" ", b""])

    def test_formfeed(self):
        """Test that form feeds are not treated as whitespace."""
        m = self.p.parse(b"!baz \fa\fb\f")
        self.assertEqual(m.arguments, [b"\fa\fb\f"])

    def test_message_ids(self):
        """Test that messages with message ids are parsed as expected."""
        m = self.p.parse(b"?bar[123]")
        self.assertEqual(m.mtype, m.REQUEST)
        self.assertEqual(m.name, "bar")
        self.assertEqual(m.arguments, [])
        self.assertEqual(m.mid, b"123")

        m = self.p.parse(b"!baz[1234] a b c")
        self.assertEqual(m.mtype, m.REPLY)
        self.assertEqual(m.name, "baz")
        self.assertEqual(m.arguments, [b"a", b"b", b"c"])
        self.assertEqual(m.mid, b"1234")

        # TODO (AJ) update katcp-python to enforce limits in the KATCP spec
        # out of range
        #self.assertRaises(KatcpSyntaxError, self.p.parse, b"!ok[0]")
        #self.assertRaises(KatcpSyntaxError, self.p.parse, b"!ok[1000000000000]")

        # bad format
        self.assertRaises(KatcpSyntaxError, self.p.parse, b"!ok[10")
        self.assertRaises(KatcpSyntaxError, self.p.parse, b"!ok[a]")

    def test_message_argument_formatting(self):
        float_val = 2.35532342334233294e17
        m = Message.request(
            'req-name', 1, float_val, True, False, 'string')
        self.assertEqual(m.arguments,
                         [b'1', repr(float_val).encode('ascii'), b'1', b'0', b'string'])

    def test_binary_data_arguments(self):
        m = self.p.parse(b'?test message \\0\\n\\r\\t\\e\\_binary')
        self.assertEqual(m, Message.request('test', b'message', b'\0\n\r\t\x1b binary'))

    def test_unicode_message_handling(self):
        unicode_str = u'!baz[1] Kl\xc3\xbcf skr\xc3\xa4m'
        self.assertRaises(KatcpTypeError, self.p.parse, unicode_str)


class TestProtocolFlags(unittest.TestCase):
    def test_parse_version(self):
        PF = katcp.ProtocolFlags
        self.assertEqual(PF.parse_version(b"foo"), PF(None, None, set()))
        self.assertEqual(PF.parse_version(b"1.0"), PF(1, 0, set()))
        self.assertEqual(
            PF.parse_version(b"5.0-MI"),
            PF(5, 0, {PF.MULTI_CLIENT, PF.MESSAGE_IDS})
        )
        # check an unknown flag
        self.assertEqual(
            PF.parse_version(b"5.1-MIU"),
            PF(5, 1, {PF.MULTI_CLIENT, PF.MESSAGE_IDS, b"U"})
        )
        # Check request timeout hint flag
        self.assertEqual(
            PF.parse_version(b"5.1-MTI"),
            PF(5, 1, {PF.MULTI_CLIENT, PF.MESSAGE_IDS, PF.REQUEST_TIMEOUT_HINTS})
        )
        # Check bulk sensor sampling flag
        self.assertEqual(
            PF.parse_version(b"5.1-MIB"),
            PF(5, 1, {PF.MULTI_CLIENT, PF.MESSAGE_IDS, PF.BULK_SET_SENSOR_SAMPLING})
        )

    def test_str(self):
        PF = katcp.ProtocolFlags
        self.assertEqual(str(PF(1, 0, set())), "1.0")
        self.assertEqual(
            str(PF(5, 0, {PF.MULTI_CLIENT, PF.MESSAGE_IDS})),
            "5.0-IM"
        )
        self.assertEqual(
            str(PF(5, 0, {PF.MULTI_CLIENT, PF.MESSAGE_IDS, b"U"})),
            "5.0-IMU"
        )
        self.assertEqual(
            str(PF(5, 1, {PF.MULTI_CLIENT, PF.MESSAGE_IDS, PF.REQUEST_TIMEOUT_HINTS})),
            "5.1-IMT"
        )
        self.assertEqual(
            str(PF(5, 1, {PF.MULTI_CLIENT, PF.MESSAGE_IDS, PF.BULK_SET_SENSOR_SAMPLING})),
            "5.1-BIM"
        )

    def test_incompatible_options(self):
        PF = katcp.ProtocolFlags
        # Katcp v4 and below don't support message ids
        with self.assertRaises(ValueError):
            PF(4, 0, [PF.MESSAGE_IDS])

        # Katcp v5.0 and below don't support timeout hint flag
        with self.assertRaises(ValueError):
            PF(5, 0, [PF.REQUEST_TIMEOUT_HINTS])

        # Katcp v5.0 and below don't support bulk sensor sampling flag
        with self.assertRaises(ValueError):
            PF(5, 0, [PF.BULK_SET_SENSOR_SAMPLING])

    def test_supports(self):
        PF = katcp.ProtocolFlags
        DUT = PF(5, 0, {b"A", b"Z"})
        self.assertTrue(DUT.supports(b"A"))
        self.assertTrue(DUT.supports(b"Z"))
        self.assertFalse(DUT.supports(b"X"))


class TestSensor(unittest.TestCase):

    def test_default_descriptions(self):
        s = Sensor(Sensor.INTEGER, 'a sens', params=[0, 10])
        self.assertEqual(s.description, "Integer sensor 'a sens' with no unit")
        s = Sensor(Sensor.FLOAT, 'fsens', None, 'microseconds', params=[0, 10])
        self.assertEqual(s.description,
                         "Float sensor 'fsens' in unit microseconds")

    def test_units_none(self):
        """Test units initialised to None results in empty string."""
        s = Sensor.integer("a sens", None, None, None)
        self.assertEqual(s.units, "")

    def test_int_sensor_from_byte_strings(self):
        """Test integer sensor initialised from byte strings."""
        s = Sensor.integer(b"an.int", b"An integer.", b"count", [-4, 3])
        # sensor attributes must be native strings
        self.assertEqual(s.name, "an.int")
        self.assertEqual(s.units, "count")
        self.assertEqual(s.description, "An integer.")
        self.assertEqual(s.stype, "integer")
        s.set(timestamp=12345, status=Sensor.NOMINAL, value=3)
        self.assertEqual(s.value(), 3)
        # after formatting, byte strings are required
        self.assertEqual(s.format_reading(s.read()), (b"12345.000000", b"nominal", b"3"))
        self.assertEqual(s.read_formatted(), (b"12345.000000", b"nominal", b"3"))

    def test_int_sensor_from_native_strings(self):
        """Test integer sensor initialised from native strings."""
        s = Sensor.integer("an.int", "An integer.", "count", [-4, 3])
        # sensor attributes must be native strings
        self.assertEqual(s.name, "an.int")
        self.assertEqual(s.units, "count")
        self.assertEqual(s.description, "An integer.")
        self.assertEqual(s.stype, "integer")
        s.set(timestamp=12345, status=Sensor.NOMINAL, value=3)
        # test both read_formatted and format_reading
        self.assertEqual(s.format_reading(s.read()), (b"12345.000000", b"nominal", b"3"))
        self.assertEqual(s.read_formatted(), (b"12345.000000", b"nominal", b"3"))
        self.assertEqual(s.parse_value(b"3"), 3)
        self.assertEqual(s.parse_value(b"4"), 4)
        self.assertEqual(s.parse_value(b"-10"), -10)
        self.assertRaises(ValueError, s.parse_value, b"asd")
        self.assertRaises(KatcpTypeError, s.parse_value, u"asd")
        self.assertRaises(KatcpTypeError, s.parse_value, u"3")

        s = Sensor(Sensor.INTEGER, "an.int", "An integer.", "count", [-20, 20])
        self.assertEqual(s.value(), 0)
        s = Sensor(Sensor.INTEGER, "an.int", "An integer.", "count", [2, 20])
        self.assertEqual(s.value(), 2)
        s = Sensor.integer("an.int", "An integer.", "count", [2, 20], default=5)
        self.assertEqual(s.value(), 5)
        self.assertEqual(s.status(), Sensor.UNKNOWN)
        s = Sensor.integer("an.int", "An integer.", "count", [2, 20],
                           initial_status=Sensor.NOMINAL)
        self.assertEqual(s.status(), Sensor.NOMINAL)

    def test_float_sensor(self):
        """Test float sensor."""
        s = Sensor.float("a.float", "A float.", "power", [0.0, 5.0])
        self.assertEqual(s.stype, 'float')
        s.set(timestamp=12345, status=Sensor.WARN, value=3.0)
        # test both read_formatted and format_reading
        self.assertEqual(s.format_reading(s.read()), (b"12345.000000", b"warn", b"3.0"))
        self.assertEqual(s.read_formatted(), (b"12345.000000", b"warn", b"3.0"))
        self.assertEqual(s.parse_value(b"3.0"), 3.0)
        self.assertEqual(s.parse_value(b"10"), 10.0)
        self.assertEqual(s.parse_value(b"-10"), -10.0)
        self.assertRaises(ValueError, s.parse_value, b"asd")

        s = Sensor(Sensor.FLOAT, "a.float", "A float.", "", [-20.0, 20.0])
        self.assertEqual(s.value(), 0.0)
        s = Sensor(Sensor.FLOAT, "a.float", "A float.", "", [2.0, 20.0])
        self.assertEqual(s.value(), 2.0)
        s = Sensor.float("a.float", "A float.", "", [2.0, 20.0], default=5.0)
        self.assertEqual(s.value(), 5.0)
        self.assertEqual(s.status(), Sensor.UNKNOWN)
        s = Sensor.float("a.float", "A float.", "", [2.0, 20.0],
                         initial_status=Sensor.WARN)
        self.assertEqual(s.status(), Sensor.WARN)

        # TODO (AJ) Test for non-ASCII fields
        # with self.assertRaises(KatcpTypeError):
        #     s = Sensor.float("a.temp", "Non-ASCII unit not allowed", "°C", default=22.0)

    def test_boolean_sensor(self):
        """Test boolean sensor."""
        s = Sensor.boolean("a.boolean", "A boolean.", "on/off", None)
        self.assertEqual(s.stype, 'boolean')
        s.set(timestamp=12345, status=Sensor.UNKNOWN, value=True)
        # test both read_formatted and format_reading
        self.assertEqual(s.format_reading(s.read()), (b"12345.000000", b"unknown", b"1"))
        self.assertEqual(s.read_formatted(), (b"12345.000000", b"unknown", b"1"))
        self.assertEqual(s.parse_value(b"1"), True)
        self.assertEqual(s.parse_value(b"0"), False)
        self.assertRaises(ValueError, s.parse_value, b"asd")
        s = Sensor.boolean("a.boolean", "A boolean.", "on/off", default=True)
        self.assertEqual(s._value, True)
        s = Sensor.boolean("a.boolean", "A boolean.", "on/off", default=False)
        self.assertEqual(s._value, False)
        s = Sensor.boolean("a.boolean", "A boolean.", "on/off",
                           initial_status=Sensor.ERROR)
        self.assertEqual(s.status(), Sensor.ERROR)

    def test_discrete_sensor_from_byte_strings(self):
        """Test discrete sensor initialised from byte strings."""
        s = Sensor.discrete(
            b"a.discrete", b"A discrete sensor.", b"state", [b"on", b"off"])
        # sensor attributes must be native strings
        self.assertEqual(s.name, "a.discrete")
        self.assertEqual(s.units, "state")
        self.assertEqual(s.description, "A discrete sensor.")
        self.assertEqual(s.stype, "discrete")
        s.set(timestamp=12345, status=Sensor.ERROR, value="on")
        self.assertEqual(s.value(), "on")
        # after formatting, byte strings are required
        self.assertEqual(s.format_reading(s.read()), (b"12345.000000", b"error", b"on"))
        self.assertEqual(s.read_formatted(), (b"12345.000000", b"error", b"on"))

    def test_discrete_sensor_from_native_strings(self):
        """Test discrete sensor initialised from native strings."""
        s = Sensor.discrete(
            "a.discrete", "A discrete sensor.", "state", ["on", "off"])
        self.assertEqual(s.stype, 'discrete')
        s.set(timestamp=12345, status=Sensor.ERROR, value="on")
        # test both read_formatted and format_reading
        self.assertEqual(s.format_reading(s.read()), (b"12345.000000", b"error", b"on"))
        self.assertEqual(s.read_formatted(), (b"12345.000000", b"error", b"on"))
        self.assertEqual(s.parse_value(b"on"), "on")
        self.assertRaises(ValueError, s.parse_value, b"fish")
        s = Sensor.discrete("a.discrete", "A discrete sensor.", "state",
                            ["on", "off"], default='on')
        self.assertEqual(s._value, 'on')
        s = Sensor.discrete("a.discrete", "A discrete sensor.", "state",
                            ["on", "off"], default='off')
        self.assertEqual(s._value, 'off')
        s = Sensor.discrete("a.discrete", "A discrete sensor.", "state",
                            ["on", "off"], initial_status=Sensor.UNREACHABLE)
        self.assertEqual(s.status(), Sensor.UNREACHABLE)

    def test_lru_sensor(self):
        """Test LRU sensor."""
        s = Sensor.lru("an.lru", "An LRU sensor.", "state", None)
        self.assertEqual(s.stype, 'lru')
        s.set(timestamp=12345, status=Sensor.FAILURE, value=Sensor.LRU_ERROR)
        # test both read_formatted and format_reading
        self.assertEqual(
            s.format_reading(s.read()), (b"12345.000000", b"failure", b"error"))
        self.assertEqual(s.read_formatted(), (b"12345.000000", b"failure", b"error"))
        self.assertEqual(s.parse_value(b"nominal"), Sensor.LRU_NOMINAL)
        self.assertRaises(ValueError, s.parse_value, b"fish")
        s = Sensor.lru(
            "an.lru", "An LRU sensor.", "state", default=Sensor.LRU_ERROR)
        self.assertEqual(s._value, Sensor.LRU_ERROR)
        s = Sensor.lru(
            "an.lru", "An LRU sensor.", "state", default=Sensor.LRU_NOMINAL)
        self.assertEqual(s._value, Sensor.LRU_NOMINAL)
        s = Sensor.lru(
            "an.lru", "An LRU sensor.", "state", initial_status=Sensor.FAILURE)
        self.assertEqual(s.status(), Sensor.FAILURE)

    def test_string_sensor(self):
        """Test string sensor."""
        s = Sensor.string("a.string", "A string sensor.", "filename", None)
        self.assertEqual(s.stype, 'string')
        s.set(timestamp=12345, status=Sensor.NOMINAL, value="zwoop")
        self.assertEqual(s._value, "zwoop")
        # test both read_formatted and format_reading
        self.assertEqual(
            s.format_reading(s.read()), (b"12345.000000", b"nominal", b"zwoop"))
        self.assertEqual(s.read_formatted(), (b"12345.000000", b"nominal", b"zwoop"))
        # test parsing byte string returns native string
        self.assertEqual(s.parse_value(b"bar foo"), "bar foo")
        # test default value
        s = Sensor.string(
            "a.string", "A string sensor.", "filename", default='baz')
        self.assertEqual(s._value, 'baz')
        # test initial status
        s = Sensor.string("a.string", "A string sensor.", "filename",
                          initial_status=Sensor.WARN)
        self.assertEqual(s.status(), Sensor.WARN)

        # Test using non-string types from set_value are "stringified"
        s = Sensor.string("a.string", "Value from list.")
        s.set_value([1, 2, 3])
        self.assertEqual(s.value(), [1, 2, 3])
        _, _, formatted_value = s.read_formatted()
        self.assertEqual(formatted_value, b'[1, 2, 3]')
        # But this is "one-way", if set from formatted byte string, the
        # "original" type is not inferred
        s.set_formatted(b'1.23', b'nominal', b'[4, 5, 6]')
        self.assertEqual(s.value(), '[4, 5, 6]')

        # Test Python version-specific limitations
        if future.utils.PY2:
            test_cases = [
                # name,       input,       raw,           from raw
                ('non-ascii', b"\x00\xff", b"\x00\xff",   b"\x00\xff"),
                ('utf-8',     u"räm",      b"r\xc3\xa4m", b"r\xc3\xa4m"),
            ]
        else:
            test_cases = [
                # name,       input,       raw,                from raw
                ('non-ascii', b"\x00\xff", b"\x00\xff",        UnicodeDecodeError),
                ('utf-8',     u"räm",      b"r\xc3\xa4m",      u"räm"),
                ('invalid',   u"\ud83d",   UnicodeEncodeError, None),
            ]
        for name, input_, raw, from_raw in test_cases:
            # set from input, then try to read back raw
            s.set_value(input_)
            if type(raw) is type and issubclass(raw, Exception):
                with self.assertRaises(raw):
                    s.read_formatted()
                continue  # cannot test setting from raw, so move on
            else:
                _, _, formatted_value = s.read_formatted()
                self.assertEqual(raw, formatted_value)

            # set raw, and try to read back value
            s.set_value(None)
            if type(from_raw) is type and issubclass(from_raw, Exception):
                with self.assertRaises(from_raw):
                    s.set_formatted(b'1.23', b'nominal', raw)
            else:
                s.set_formatted(b'1.23', b'nominal', raw)
                self.assertEqual(s.value(), from_raw)

    def test_timestamp_sensor(self):
        """Test timestamp sensor."""
        s = Sensor.timestamp("a.timestamp", "A timestamp sensor.", "", None)
        self.assertEqual(s.stype, 'timestamp')
        s.set(timestamp=12345, status=Sensor.NOMINAL, value=1001.9)
        # test both read_formatted and format_reading
        self.assertEqual(s.format_reading(s.read()),
                         (b"12345.000000", b"nominal", b"1001.900000"))
        self.assertEqual(s.read_formatted(),
                         (b"12345.000000", b"nominal", b"1001.900000"))
        # Test with katcp v4 parsing formatting
        self.assertEqual(s.format_reading(s.read(), major=4),
                         (b"12345000", b"nominal", b"1001900"))
        self.assertEqual(s.read_formatted(major=4),
                         (b"12345000", b"nominal", b"1001900"))
        self.assertAlmostEqual(s.parse_value(b"1002.100"), 1002.1)
        self.assertRaises(ValueError, s.parse_value, b"bicycle")
        s = Sensor.timestamp(
            "a.timestamp", "A timestamp sensor.", "", default=123)
        self.assertEqual(s._value, 123)

        s.set_formatted(b'12346.1', b'nominal', b'12246.1')
        self.assertEqual(s.read(), (12346.1, Sensor.NOMINAL, 12246.1))

        # Test with katcp v4 parsing
        s.set_formatted(b'12347100', b'nominal', b'12247100', major=4)
        self.assertEqual(s.read(), (12347.1, Sensor.NOMINAL, 12247.1))

        s = Sensor.timestamp("a.timestamp", "A timestamp sensor.", "",
                             initial_status=Sensor.NOMINAL)
        self.assertEqual(s.status(), Sensor.NOMINAL)

    def test_address_sensor(self):
        """Test address sensor."""
        s = Sensor.address("a.address", "An address sensor.", "", None)
        self.assertEqual(s.stype, 'address')
        s.set(timestamp=12345, status=Sensor.NOMINAL, value=("127.0.0.1", 80))
        # test both read_formatted and format_reading
        self.assertEqual(s.format_reading(s.read()),
                         (b"12345.000000", b"nominal", b"127.0.0.1:80"))
        self.assertEqual(s.read_formatted(),
                         (b"12345.000000", b"nominal", b"127.0.0.1:80"))
        self.assertEqual(s.parse_value(b"[::1]:80"), ("::1", 80))
        self.assertRaises(ValueError, s.parse_value, b"[::1]:foo")
        s = Sensor.address("a.address", "An address sensor.", "",
                           default=('192.168.101.1', 81))
        self.assertEqual(s._value, ('192.168.101.1', 81))
        s = Sensor.address("a.address", "An address sensor.", "",
                           initial_status=Sensor.NOMINAL)
        self.assertEqual(s.status(), Sensor.NOMINAL)

    def test_set_and_get_value(self):
        """Test getting and setting a sensor value."""
        s = Sensor.integer("an.int", "An integer.", "count", [-4, 3])
        s.set(timestamp=12345, status=Sensor.NOMINAL, value=3)

        self.assertEqual(s.value(), 3)

        s.set_value(2)
        self.assertEqual(s.value(), 2)

        s.set_value(3, timestamp=12345)
        self.assertEqual(s.read(), (12345, Sensor.NOMINAL, 3))

        s.set_value(5, timestamp=12345)
        self.assertEqual(s.read(), (12345, Sensor.NOMINAL, 5))

        s.set_formatted(b'12346.1', b'nominal', b'-2')
        self.assertEqual(s.read(), (12346.1, Sensor.NOMINAL, -2))

        # Test setting with katcp v4 parsing
        s.set_formatted(b'12347100', b'warn', b'-3', major=4)
        self.assertEqual(s.read(), (12347.1, Sensor.WARN, -3))

    def test_statuses(self):
        # Test that the status constants are all good
        valid_statuses = set(['unknown', 'nominal', 'warn', 'error',
                              'failure', 'unreachable', 'inactive'])
        status_vals_set = set()
        status_vals_dict = {}
        for st in valid_statuses:
            st_byte_str = st.encode('ascii')
            # Check that a capitalised attribute exists for each status
            st_val = getattr(Sensor, st.upper(), 'OOPS')
            self.assertNotEqual(st_val, 'OOPS')
            # Check that the status to name lookup dict is correct
            self.assertEqual(Sensor.STATUSES[st_val], st)
            self.assertEqual(Sensor.STATUSES_RAW[st_val], st_byte_str)
            # Check that the name to value lookup dict is correct
            self.assertEqual(Sensor.STATUS_NAMES[st], st_val)
            self.assertEqual(Sensor.STATUS_NAMES_RAW[st_byte_str], st_val)
            status_vals_set.add(st_val)
            status_vals_dict[st] = st_val

        # Check that the status values are all unique
        self.assertEqual(len(status_vals_set), len(valid_statuses))
        # Check that there are not extra entries in the forward/backward name
        # lookup dicts
        self.assertEqual(len(Sensor.STATUSES), len(valid_statuses))
        self.assertEqual(len(Sensor.STATUS_NAMES), len(valid_statuses))


class TestAsyncState(tornado.testing.AsyncTestCase):

    def setUp(self):
        super(TestAsyncState, self).setUp()
        self._valid_states = ['on', 'off', 'floating']
        self._state = AsyncState(self._valid_states, 'off', self.io_loop)

    def test_init(self):
        self.assertEqual(self._state.state, 'off')
        self.assertEqual(sorted(self._state.valid_states),
                         sorted(self._valid_states))

    @tornado.testing.gen_test
    def test_timeout_of_until_state(self):
        @tornado.gen.coroutine
        def set_state_on():
            self._state.set_state('on')
        # Test for timing out
        with self.assertRaises(tornado.gen.TimeoutError):
            yield self._state.until_state('on', timeout=0.1)
        # Test for NOT timing out
        self.io_loop.add_callback(set_state_on)
        yield self._state.until_state('on', timeout=0.1)

    @tornado.testing.gen_test
    def test_timeout_of_until_state_in(self):
        @tornado.gen.coroutine
        def set_state_floating():
            self._state.set_state('floating')
        # Test for timing out
        with self.assertRaises(tornado.gen.TimeoutError):
            yield self._state.until_state_in('on', 'floating', timeout=0.1)

        # Test for NOT timing out
        self.io_loop.add_callback(set_state_floating)
        yield self._state.until_state_in('on', 'floating', timeout=0.1)


class TestAsyncEvent(tornado.testing.AsyncTestCase):

    def setUp(self):
        super(TestAsyncEvent, self).setUp()
        self._event = AsyncEvent(self.io_loop)

    @tornado.testing.gen_test
    def test_timeout_of_until_set(self):
        @tornado.gen.coroutine
        def call_set():
            self._event.set()
        # Test for timing out
        with self.assertRaises(tornado.gen.TimeoutError):
            yield self._event.until_set(timeout=0.1)
        # Test for NOT timing out
        self.io_loop.add_callback(call_set)
        yield self._event.until_set(timeout=0.1)


class TestUntilSome(tornado.testing.AsyncTestCase):

    @tornado.testing.gen_test
    def test_until_some_args(self):
        results = yield until_some()
        self.assertEqual(results, [], 'Expected empty list for until_some()')
        f1 = tornado.concurrent.Future()
        f2 = tornado.concurrent.Future()
        f3 = tornado.concurrent.Future()
        # Test for timing out
        with self.assertRaises(tornado.gen.TimeoutError):
            yield until_some(f1, f2, f3, timeout=0.05)
        # Test for NOT timing out
        f3.set_result(84)
        f1.set_result(24)
        f2.set_result(42)
        results = yield until_some(f1, f2, f3, timeout=0.1)
        self.assertEqual(sorted(results), [(0, 24), (1, 42), (2, 84)],
                         'Results differ for until_some (3 arg futures)')

    @tornado.testing.gen_test
    def test_until_some_kwargs(self):
        f1 = tornado.concurrent.Future()
        f2 = tornado.concurrent.Future()
        f3 = tornado.concurrent.Future()
        futures = {'f1': f1, 'f2': f2, 'f3': f3}
        f1.set_result(24)
        with self.assertRaises(tornado.gen.TimeoutError):
            yield until_some(done_at_least=2, timeout=0.05, **futures)
        f2.set_result(42)
        results = yield until_some(done_at_least=1, timeout=0.1, **futures)
        # both f1 and f2 are ready, but kwargs dict makes order unpredictable
        self.assertEqual(len(results), 1)
        options = {'f1': 24, 'f2': 42}
        self.assertDictContainsSubset(dict(results), options,
                                      'Results differ for until_some (1 kwarg future)')
        f3.set_result(84)
        results = yield until_some(done_at_least=2, timeout=0.1, **futures)
        # similar to above, any 2 of the 3 kwarg futures could be returned
        self.assertEqual(len(results), 2)
        options = {'f1': 24, 'f2': 42, 'f3': 84}
        self.assertDictContainsSubset(dict(results), options,
                                      'Results differ for until_some (2 kwarg futures)')


class TestHashableIdentity(unittest.TestCase):

    def test_hashable_identity(self):

        class MyObject(object):
            def my_method(self):
                pass

        def my_func1():
            pass

        def my_func2():
            pass

        obj1 = MyObject()
        obj2 = MyObject()

        hash1 = hashable_identity(obj1)
        hash1_again = hashable_identity(obj1)
        hash2 = hashable_identity(obj2)
        self.assertEqual(hash1, hash1_again)
        self.assertNotEqual(hash1, hash2)

        hash1 = hashable_identity(obj1.my_method)
        hash1_again = hashable_identity(obj1.my_method)
        hash2 = hashable_identity(obj2.my_method)
        self.assertEqual(hash1, hash1_again)
        self.assertNotEqual(hash1, hash2)

        hash1 = hashable_identity(my_func1)
        hash1_again = hashable_identity(my_func1)
        hash2 = hashable_identity(my_func2)
        self.assertEqual(hash1, hash1_again)
        self.assertNotEqual(hash1, hash2)

        hash1 = hashable_identity('baz')
        hash1_again = hashable_identity('baz')
        hash2 = hashable_identity('foo')
        self.assertEqual(hash1, hash1_again)
        self.assertNotEqual(hash1, hash2)



# Copied from peak.util.proxies https://pypi.python.org/pypi/ProxyTypes/0.9
# to allow translation to python 2/3 compatible code # According to the pypi
# index record this code is distributed under PSF (Python Software License) or
# ZPL (Zope Public License if I' not mistaken). The PSF license is BSD style and
# allows us to copy, modify and redistribute this code.

from __future__ import absolute_import, division, print_function
from future import standard_library
standard_library.install_aliases()  # noqa: E402

from builtins import object


class AbstractProxy(object):
    """Delegates all operations (except ``.__subject__``) to another object"""
    __slots__ = ()

    def __call__(self,*args,**kw):
        return self.__subject__(*args,**kw)

    def __getattribute__(self, attr, oga=object.__getattribute__):
        subject = oga(self,'__subject__')
        if attr=='__subject__':
            return subject
        return getattr(subject,attr)

    def __setattr__(self,attr,val, osa=object.__setattr__):
        if attr=='__subject__':
            osa(self,attr,val)
        else:
            setattr(self.__subject__,attr,val)

    def __delattr__(self,attr, oda=object.__delattr__):
        if attr=='__subject__':
            oda(self,attr)
        else:
            delattr(self.__subject__,attr)

    def __bool__(self):
        return bool(self.__subject__)

    def __getitem__(self,arg):
        return self.__subject__[arg]

    def __setitem__(self,arg,val):
        self.__subject__[arg] = val

    def __delitem__(self,arg):
        del self.__subject__[arg]

    def __getslice__(self,i,j):
        return self.__subject__[i:j]


    def __setslice__(self,i,j,val):
        self.__subject__[i:j] = val

    def __delslice__(self,i,j):
        del self.__subject__[i:j]

    def __contains__(self,ob):
        return ob in self.__subject__

    for name in 'repr str hash len abs complex int long float iter oct hex'.split():
        exec("def __%s__(self): return %s(self.__subject__)" % (name,name))

    for name in 'cmp', 'coerce', 'divmod':
        exec("def __%s__(self,ob): return %s(self.__subject__,ob)" % (name,name))

    for name,op in [
        ('lt','<'), ('gt','>'), ('le','<='), ('ge','>='),
        ('eq','=='), ('ne','!=')
    ]:
        exec("def __%s__(self,ob): return self.__subject__ %s ob" % (name,op))

    for name,op in [('neg','-'), ('pos','+'), ('invert','~')]:
        exec("def __%s__(self): return %s self.__subject__" % (name,op))

    for name, op in [
        ('or','|'),  ('and','&'), ('xor','^'), ('lshift','<<'), ('rshift','>>'),
        ('add','+'), ('sub','-'), ('mul','*'), ('div','/'), ('mod','%'),
        ('truediv','/'), ('floordiv','//')
    ]:
        exec((
            "def __%(name)s__(self,ob):\n"
            "    return self.__subject__ %(op)s ob\n"
            "\n"
            "def __r%(name)s__(self,ob):\n"
            "    return ob %(op)s self.__subject__\n"
            "\n"
            "def __i%(name)s__(self,ob):\n"
            "    self.__subject__ %(op)s=ob\n"
            "    return self\n"
        )  % locals())

    del name, op

    # Oddball signatures

    def __rdivmod__(self,ob):
        return divmod(ob, self.__subject__)

    def __pow__(self,*args):
        return pow(self.__subject__,*args)

    def __ipow__(self,ob):
        self.__subject__ **= ob
        return self

    def __rpow__(self,ob):
        return pow(ob, self.__subject__)


class ObjectProxy(AbstractProxy):
    """Proxy for a specific object"""

    __slots__ = "__subject__"

    def __init__(self,subject):
        self.__subject__ = subject


class CallbackProxy(AbstractProxy):
    """Proxy for a dynamically-chosen object"""

    __slots__ = '__callback__'

    def __init__(self, func):
        set_callback(self,func)

set_callback = CallbackProxy.__callback__.__set__
get_callback = CallbackProxy.__callback__.__get__
CallbackProxy.__subject__ = property(lambda self, gc=get_callback: gc(self)())



class LazyProxy(CallbackProxy):
    """Proxy for a lazily-obtained object, that is cached on first use"""
    __slots__ = "__cache__"

get_cache = LazyProxy.__cache__.__get__
set_cache = LazyProxy.__cache__.__set__

def __subject__(self, get_cache=get_cache, set_cache=set_cache):
    try:
        return get_cache(self)
    except AttributeError:
        set_cache(self, get_callback(self)())
        return get_cache(self)

LazyProxy.__subject__ = property(__subject__, set_cache)
del __subject__


class AbstractWrapper(AbstractProxy):
    """Mixin to allow extra behaviours and attributes on proxy instance"""
    __slots__ = ()

    def __getattribute__(self, attr, oga=object.__getattribute__):
        if attr.startswith('__'):
            subject = oga(self,'__subject__')
            if attr=='__subject__':
                return subject
            return getattr(subject,attr)
        return oga(self,attr)

    def __getattr__(self,attr, oga=object.__getattribute__):
        return getattr(oga(self,'__subject__'), attr)

    def __setattr__(self,attr,val, osa=object.__setattr__):
        if (
            attr=='__subject__'
            or hasattr(type(self),attr) and not attr.startswith('__')
        ):
            osa(self,attr,val)
        else:
            setattr(self.__subject__,attr,val)

    def __delattr__(self,attr, oda=object.__delattr__):
        if (
            attr=='__subject__'
            or hasattr(type(self),attr) and not attr.startswith('__')
        ):
            oda(self,attr)
        else:
            delattr(self.__subject__,attr)

class ObjectWrapper(ObjectProxy, AbstractWrapper):      __slots__ = ()
class CallbackWrapper(CallbackProxy, AbstractWrapper):  __slots__ = ()
class LazyWrapper(LazyProxy, AbstractWrapper):          __slots__ = ()



# __init__.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2009 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

"""Tests for the katcp package."""

# Automatically added by katversion
__version__ = '0.9.0'



# Copyright 2010 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

from __future__ import absolute_import, division, print_function

import sys

from katcp import DeviceServer, Message, Sensor
from util import standard_parser


class BenchmarkServer(DeviceServer):
    # an ugly hack
    def _bind(self, *args):
        res = DeviceServer._bind(self, *args)
        print("RUNNING")
        sys.stdout.flush()
        return res

    def setup_sensors(self):
        pass

    def request_add_sensor(self, sock, msg):
        """ add a sensor
        """
        self.add_sensor(Sensor(int, 'int_sensor%d' % len(self._sensors),
                               'descr', 'unit', params=[-10, 10]))
        return Message.reply('add-sensor', 'ok')

def main():
    parser = standard_parser(1236)
    options, args = parser.parse_args()
    server = BenchmarkServer('localhost', options.port)
    server.run()

if __name__ == '__main__':
    main()



# Copyright 2010 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

"""Sensor tree implementation.

A sensor tree is a DAG (directed acyclic graph) of sensor objects
where an edge represents a dependency. E.g.

S1 -> S2
  |
  -> S3 -> S4
        |
S6 -------> S5

is a sensor tree where the value of S3 depends on the values of S4
and S5, the value of S1 depends on S3 and S2 and the value of S6
depends on just S5.

When a sensor is added to the tree, the tree attaches itself to the
sensor's update notification list. A sensor update triggers a recalculation
of the sensor values that depend on it. These value changes may then trigger
further updates.

The acyclic requirement on the graph structure is required to ensure
that the update chain eventually terminates. It is not enforced.

"""
from __future__ import absolute_import, division, print_function
from future import standard_library
standard_library.install_aliases()  # noqa: E402

from builtins import object


class GenericSensorTree(object):
    """A tree of generic sensors."""
    # TODO: consider adding detection of cycles.
    #       Fastest method is probably maintaining a spanning tree
    #       and checking for back edges. As always, deletions are likely
    #       to be a bit of a pain and require tree shuffling.
    #       Disconnected graphs will also make life a little more difficult.
    def __init__(self):
        # map of child -> set of all parent sensors
        self._child_to_parents = {}
        # map of parent -> set of all child sensors
        self._parent_to_children = {}

    def update(self, sensor, reading):
        """Update callback used by sensors to notify observers of changes.

        Parameters
        ----------
        sensor : :class:`katcp.Sensor` object
            The sensor whose value has changed.
        reading : (timestamp, status, value) tuple
            Sensor reading as would be returned by sensor.read()

        """
        parents = list(self._child_to_parents[sensor])
        for parent in parents:
            self.recalculate(parent, (sensor,))

    def recalculate(self, parent, updates):
        """Re-calculate the value of parent sensor.

        Sub-classes should override this method and call parent.set_value(...)
        with the new parent sensor value.

        Recalculate is called with a single child sensor when a sensor value
        is updated. It is called by add_links and remove_links with the same
        list of children they were called with when once links have been
        added or removed.

        Parameters
        ----------
        parent : :class:`katcp.Sensor` object
            The sensor that needs to be updated.
        updates : sequence of :class:`katcp.Sensor` objects
            The child sensors which triggered the update.

        """
        raise NotImplementedError

    def _add_sensor(self, sensor):
        """Add a new sensor to the tree.

        Parameters
        ----------
        sensor : :class:`katcp.Sensor` object
            New sensor to add to the tree.

        """
        self._parent_to_children[sensor] = set()
        self._child_to_parents[sensor] = set()

    def _remove_sensor(self, sensor):
        """Remove a sensor from the tree.

        Parameters
        ----------
        sensor : :class:`katcp.Sensor` object
            Sensor to remove from the tree.

        """
        del self._parent_to_children[sensor]
        del self._child_to_parents[sensor]

    def add_links(self, parent, children):
        """Create dependency links from parent to child.

        Any sensors not in the tree are added. After all dependency links have
        been created, the parent is recalculated and the tree attaches to any
        sensors it was not yet attached to. Links that already exist are
        ignored.

        Parameters
        ----------
        parent : :class:`katcp.Sensor` object
            The sensor that depends on children.
        children : sequence of :class:`katcp.Sensor` objects
            The sensors parent depends on.

        """
        new_sensors = []
        if parent not in self:
            self._add_sensor(parent)
            new_sensors.append(parent)
        for child in children:
            if child not in self:
                self._add_sensor(child)
                new_sensors.append(child)
            self._parent_to_children[parent].add(child)
            self._child_to_parents[child].add(parent)

        self.recalculate(parent, children)
        for sensor in new_sensors:
            sensor.attach(self)

    def remove_links(self, parent, children):
        """Remove dependency links from parent to child.

        Any sensors that have no dependency links are removed from the tree and
        the tree detaches from each sensor removed. After all dependency links
        have been removed the parent is recalculated. Links that don't exist
        are ignored.

        Parameters
        ----------
        parent : :class:`katcp.Sensor` object
            The sensor that used to depend on children.
        children : sequence of :class:`katcp.Sensor` objects
            The sensors that parent used to depend on.

        """
        old_sensors = []
        if parent in self:
            for child in children:
                if child not in self:
                    continue
                self._parent_to_children[parent].discard(child)
                self._child_to_parents[child].discard(parent)
                if not self._child_to_parents[child] and \
                        not self._parent_to_children[child]:
                    self._remove_sensor(child)
                    old_sensors.append(child)
            if not self._child_to_parents[parent] and \
                    not self._parent_to_children[parent]:
                self._remove_sensor(parent)
                old_sensors.append(parent)

        for sensor in old_sensors:
            sensor.detach(self)
        self.recalculate(parent, children)

    def children(self, parent):
        """Return set of children of parent.

        Parameters
        ----------
        parent : :class:`katcp.Sensor` object
            Parent whose children to return.

        Returns
        -------
        children : set of :class:`katcp.Sensor` objects
            The child sensors of parent.

        """
        if parent not in self._parent_to_children:
            raise ValueError("Parent sensor %r not in tree." % parent)
        return self._parent_to_children[parent].copy()

    def parents(self, child):
        """Return set of parents of child.

        Parameters
        ----------
        child : :class:`katcp.Sensor` object
            Child whose parents to return.

        Returns
        -------
        parents : set of :class:`katcp.Sensor` objects
            The parent sensors of child.

        """
        if child not in self._child_to_parents:
            raise ValueError("Child sensor %r not in tree." % child)
        return self._child_to_parents[child].copy()

    def __contains__(self, sensor):
        """Return True if sensor is in the tree, False otherwise.

        Parameters
        ----------
        sensor : object
            Sensor to check for in tree. Objects that are not sensors
            cannot appear in the tree and so will return False.

        """
        return sensor in self._parent_to_children


class BooleanSensorTree(GenericSensorTree):
    """A tree of boolean sensors.

    Non-leaf sensors have their values updated to be the logical AND
    of their child nodes.

    Examples
    --------
    >>> from katcp import Sensor, BooleanSensorTree
    >>> tree = BooleanSensorTree()
    >>> sensor1 = Sensor(Sensor.BOOLEAN, "sensor1", "First sensor", "")
    >>> sensor2 = Sensor(Sensor.BOOLEAN, "sensor2", "Second sensor", "")
    >>> tree.add(sensor1, sensor2)
    >>> sensor2.set_value(True)
    >>> sensor1.value()
    >>> sensor2.set_value(False)
    >>> sensor1.value()
    >>> tree.remove(sensor1, sensor2)
    >>> sensor1.value()

    """
    def __init__(self):
        super(BooleanSensorTree, self).__init__()
        # map of parent -> set of child sensors not ok
        # key None points to root nodes
        self._parent_to_not_ok = {}

    def add(self, parent, child):
        """Add a pair of boolean sensors.

        Parent depends on child.

        Parameters
        ----------
        parent : boolean instance of :class:`katcp.Sensor`
            The sensor that depends on child.
        child : boolean instance of :class:`katcp.Sensor`
            The sensor parent depends on.

        """
        if parent not in self:
            if parent.stype != "boolean":
                raise ValueError("Parent sensor %r is not boolean" % child)
            self._parent_to_not_ok[parent] = set()
        if child not in self:
            if child.stype != "boolean":
                raise ValueError("Child sensor %r is not boolean" % child)
            self._parent_to_not_ok[child] = set()
        self.add_links(parent, (child,))

    def remove(self, parent, child):
        """Remove a dependency between parent and child.

        Parameters
        ----------
        parent : boolean instance of :class:`katcp.Sensor`
            The sensor that used to depend on child.
        child : boolean instance of :class:`katcp.Sensor` or None
            The sensor parent used to depend on.

        """
        self.remove_links(parent, (child,))
        if parent not in self and parent in self._parent_to_not_ok:
            del self._parent_to_not_ok[parent]
        if child not in self and child in self._parent_to_not_ok:
            del self._parent_to_not_ok[child]

    def recalculate(self, parent, updates):
        """Re-calculate the value of parent sensor.

        Parent's value is the boolean AND of all child sensors.

        Parameters
        ----------
        parent : :class:`katcp.Sensor` object
            The sensor that needs to be updated.
        updates : sequence of :class:`katcp.Sensor` objects
            The child sensors which triggered the update.

        """
        not_ok = self._parent_to_not_ok[parent]
        children = self.children(parent) if parent in self else set()
        for sensor in updates:
            if sensor not in children or sensor.value():
                not_ok.discard(sensor)
            else:
                not_ok.add(sensor)
        parent.set_value(not not_ok)


class AggregateSensorTree(GenericSensorTree):
    """A collection of aggregate sensors.

    Examples
    --------

    Example where sensors are available when rules are added::

    >>> from katcp import Sensor, AggregateSensorTree
    >>> tree = AggregateSensorTree()
    >>> def add_rule(parent, children):
    >>>     parent.set_value(sum(child.value() for child in children))
    >>> sensor1 = Sensor(Sensor.INTEGER, "sensor1", "First sensor", "",
    ...                  [-1000, 1000])
    >>> sensor2 = Sensor(Sensor.INTEGER, "sensor2", "Second sensor", "",
    ...                  [-1000, 1000])
    >>> agg = Sensor(Sensor.INTEGER, "sum", "The total", "", [-2000, 2000])
    >>> tree.add(agg, add_rule, (sensor1, sensor2))
    >>> agg.value()
    >>> sensor1.set_value(1)
    >>> agg.value()
    >>> sensor2.set_value(2)
    >>> agg.value()
    >>> tree.remove(agg)
    >>> agg.value()

    Example where rules need to be added before dependent sensors are
    available::

    >>> from katcp import Sensor, AggregateSensorTree
    >>> tree = AggregateSensorTree()
    >>> def add_rule(parent, children):
    >>>     parent.set_value(sum(child.value() for child in children))
    >>> agg = Sensor(Sensor.INTEGER, "sum", "The total", "", [-2000, 2000])
    >>> tree.add_delayed(agg, add_rule, ("sensor1", "sensor2"))
    >>> agg.value()
    >>> sensor1 = Sensor(Sensor.INTEGER, "sensor1", "First sensor", "",
    ...                  [-1000, 1000])
    >>> sensor1.set_value(5)
    >>> tree.register_sensor(sensor1)
    >>> agg.value() # still 0
    >>> sensor2 = Sensor(Sensor.INTEGER, "sensor2", "Second sensor", "",
    ...                  [-1000, 1000])
    >>> sensor2.set_value(3)
    >>> tree.register_sensor(sensor2)
    >>> agg.value() # now 8

    """
    def __init__(self):
        super(AggregateSensorTree, self).__init__()
        # map of aggregate sensor -> (rule_function, children)
        self._aggregates = {}
        # map of incomplete aggregate sensor -> (rule_function, names, sensors)
        # as sensor are found they are migrated from names to sensors
        self._incomplete_aggregates = {}
        # map of sensor name -> registered sensor
        self._registered_sensors = {}

    def add(self, parent, rule_function, children):
        """Create an aggregation rule.

        Parameters
        ----------
        parent : :class:`katcp.Sensor` object
            The aggregate sensor.
        rule_function : f(parent, children)
            Function to update the parent sensor value.
        children : sequence of :class:`katcp.Sensor` objects
            The sensors the aggregate sensor depends on.

        """
        if parent in self._aggregates or parent in self._incomplete_aggregates:
            raise ValueError("Sensor %r already has an aggregate rule "
                             "associated" % parent)
        self._aggregates[parent] = (rule_function, children)
        self.add_links(parent, children)

    def add_delayed(self, parent, rule_function, child_names):
        """Create an aggregation rule before child sensors are present.

        Parameters
        ----------
        parent : :class:`katcp.Sensor` object
            The aggregate sensor.
        rule_function : f(parent, children)
            Function to update the parent sensor value.
        child_names : sequence of str
            The names of the sensors the aggregate sensor depends on. These
            sensor must be registered using :meth:`register_sensor` to become
            active.

        """
        if parent in self._aggregates or parent in self._incomplete_aggregates:
            raise ValueError("Sensor %r already has an aggregate rule"
                             " associated" % parent)
        reg = self._registered_sensors
        names = set(name for name in child_names if name not in reg)
        sensors = set(reg[name] for name in child_names if name in reg)
        if names:
            self._incomplete_aggregates[parent] = (rule_function, names,
                                                   sensors)
        else:
            self.add(parent, rule_function, sensors)

    def register_sensor(self, child):
        """Register a sensor required by an aggregate sensor registered with
        add_delayed.

        Parameters
        ----------
        child : :class:`katcp.Sensor` object
            A child sensor required by one or more delayed aggregate sensors.

        """
        child_name = self._get_sensor_reference(child)
        if child_name in self._registered_sensors:
            raise ValueError("Sensor %r already registered with aggregate"
                             " tree" % child)
        self._registered_sensors[child_name] = child
        completed = []
        for parent, (_rule, names, sensors) in self._incomplete_aggregates.items():
            if child_name in names:
                names.remove(child_name)
                sensors.add(child)
                if not names:
                    completed.append(parent)
        for parent in completed:
            rule_function, _names, sensors = \
                self._incomplete_aggregates[parent]
            del self._incomplete_aggregates[parent]
            self.add(parent, rule_function, sensors)

    def deregister_sensor(self, child):
        child_name = self._get_sensor_reference(child)
        if child_name not in self._registered_sensors:
            raise ValueError("Sensor %r not registered with aggregate"
                             " tree" % child_name)
        child = self._registered_sensors[child_name]
        del self._registered_sensors[child_name]
        try:
            parents = self.parents(child)
            for parent in parents:
                if parent in self._incomplete_aggregates:
                    (rule, names, sensors) = self._incomplete_aggregates[parent]
                    names.add(child_name)
                else:
                    (rule, sensors) = self._aggregates[parent]
                    names = set()
                    names.add(child_name)
                sensors.discard(child)
                self._incomplete_aggregates[parent] = (rule, names, sensors)
                self.remove_links(parent, [child])
                self.remove(parent)
                if parent.stype == 'bool':
                    parent.set_value(False)
        except ValueError:
            pass

    def _child_from_reference(self, reference):
        """Returns the child sensor from its reference.

        Parameters
        ----------
        reference : str
            Reference to sensor (typically its name).

        Returns
        -------
        child : :class:`katcp.Sensor` object
            A child sensor linked to one or more aggregate sensors.

        """
        for child in self._child_to_parents:
            if self._get_sensor_reference(child) == reference:
                return child

    def remove(self, parent):
        """Remove an aggregation rule.

        Parameters
        ----------
        parent : :class:`katcp.Sensor` object
            The aggregate sensor to remove.

        """
        if parent not in self._aggregates:
            raise ValueError("Sensor %r does not have an aggregate rule "
                             "associated" % parent)
        children = self.children(parent)
        try:
            self.remove_links(parent, children)
        except Exception:
            pass
        del self._aggregates[parent]

    def fetch(self, parent):
        """Retrieve an aggregation rule.

        Parameters
        ----------
        parent : :class:`katcp.Sensor` object

        Returns
        -------
        rule_function : f(parent, children)
            Function give to update the parent sensor value.
        children : sequence of :class:`katcp.Sensor` objects
            The sensors the aggregate sensor depends on.

        """
        return self._aggregates[parent]

    def recalculate(self, parent, updates):
        """Re-calculate the value of parent sensor.

        Parent's value is calculated by calling the associate aggregation rule.

        Parameters
        ----------
        parent : :class:`katcp.Sensor` object
            The sensor that needs to be updated.
        updates : sequence of :class:`katcp.Sensor` objects
            The child sensors which triggered the update.

        """
        rule_function, children = self._aggregates[parent]
        rule_function(parent, children)

    def _get_sensor_reference(self, sensor):
        """Returns sensor name as reference for sensors to be registered by.

        Parameters
        ----------
        sensor : :class:`katcp.Sensor` object
            Sensor to refer to.

        Returns
        -------
        reference : str
            Sensor name as reference.

        """
        return sensor.name
